#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Simulation manager creates and controls the simulation objects."""
import logging
from typing import Callable

import pandas as pd

from omotes_simulator_core.adapter.transforms.mappers import (
    EsdlControllerMapper,
    EsdlEnergySystemMapper,
)
from omotes_simulator_core.entities.esdl_object import EsdlObject
from omotes_simulator_core.entities.heat_network import HeatNetwork
from omotes_simulator_core.entities.simulation_configuration import SimulationConfiguration
from omotes_simulator_core.simulation import NetworkSimulation

logger = logging.getLogger(__name__)


class SimulationManager:
    """Manager class for managing the simulation."""

    def __init__(self, esdl: EsdlObject, config: SimulationConfiguration):
        """Constructor for SimulationManager class.

        :param EsdlObject esdl: Esdlobject, which stores the network information
        :param SimulationConfiguration config: Config object to hold the simulation start, stop, end
        """
        self.esdl = esdl
        self.config = config

    def execute(self, progress_calback: Callable[[float, str], None]) -> pd.DataFrame:
        """Method to simulate the network.

        First the network is converted to an internal object model and then the simulation is run.

        :return: DataFrame with the result of the simulations
        """
        # convert ESDL to Heat Network, NetworkController
        network = HeatNetwork(EsdlEnergySystemMapper(self.esdl).to_entity)
        controller = EsdlControllerMapper().to_entity(self.esdl)

        worker = NetworkSimulation(network, controller)
        worker.run(self.config, progress_calback)

        # Run output presenter that iterates over het network (/controller?) and
        # gathers the output into a single data object
        return worker.gather_output()



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Entry point for running simulator-core library from cmdline."""

import logging
import sys
import traceback
import uuid
from datetime import datetime

import pandas as pd

from omotes_simulator_core.entities.esdl_object import EsdlObject
from omotes_simulator_core.entities.simulation_configuration import SimulationConfiguration
from omotes_simulator_core.infrastructure.simulation_manager import SimulationManager
from omotes_simulator_core.infrastructure.utils import pyesdl_from_file

logger = logging.getLogger(__name__)


def progressLogger(progress: float, message: str) -> None:
    """Function to report progress to logging/stdout."""
    logger.info(f"{message} ({progress*100:.2f}%)")


def run(file_path: str | None = None) -> pd.DataFrame:
    """Main run function for the heatnetwork simulator."""
    config = SimulationConfiguration(
        simulation_id=uuid.uuid1(),
        name="test run",
        timestep=3600,
        start=datetime.strptime("2019-01-01T00:00:00", "%Y-%m-%dT%H:%M:%S"),
        stop=datetime.strptime("2019-01-01T01:00:00", "%Y-%m-%dT%H:%M:%S"),
    )

    esdl_file_path = sys.argv[1] if file_path is None else file_path
    try:
        app = SimulationManager(EsdlObject(pyesdl_from_file(esdl_file_path)), config)
        result = app.execute(progressLogger)
        return result
    except Exception as error:
        logger.error(f"Error occured: {error} at: {traceback.format_exc(limit=-1)}")
        logger.debug(traceback.format_exc())
        raise error


if __name__ == "__main__":
    # Set loglevel to logging.DEBUG for more verbose output
    logging.basicConfig(
        level=logging.INFO, format="%(asctime)s [%(levelname)s]:%(name)s - %(message)s"
    )
    t1 = datetime.now()
    result = run(r".\testdata\test1.esdl")
    t2 = datetime.now()

    logger.info(f"Results dataframe shape=({result.shape})")
    logger.info(f"Execution time: {t2-t1}")
    logger.debug(result.head())



# code copied from: https://github.com/Dennis-van-Gils/python-fluidprop/blob/main/src/fluidprop.py
#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Easy access to thermodynamic fluid properties as a function of temperature
and pressure. Comes with a minimal command-line interface for quick inspection.
Provides class ``FluidProperties()`` useful for working out dataseries in your
own scripts.

Thermodynamic properties are provided by CoolProp:
* http://www.coolprop.org/
* http://pubs.acs.org/doi/abs/10.1021/ie4033999
"""
__author__ = "Dennis van Gils"
__authoremail__ = "vangils.dennis@gmail.com"
__url__ = "https://github.com/Dennis-van-Gils/python-fluidprop"
__date__ = "13-05-2024"
__version__ = "1.2.0"

import re
from typing import Union

import numpy.typing as npt
import numpy as np
import CoolProp
import CoolProp.CoolProp as CP


ZERO_C = 273.15
"""0 'C in K"""
P_ATM = 1.01325
"""bar per 1 atm"""
P_PSI = 1 / 14.504
"""bar per 1 psi"""
HRULE = "-" * 60
"""Horizontal rule"""

# fmt: off
FLUID_SELECTION = [
    ("Air"                  , "mixture"),
    ("Hydrogen"             , "H_{2}"),
    ("Helium"               , "He"),
    ("Nitrogen"             , "N_{2}"),
    ("Oxygen"               , "O_{2}"),
    ("CarbonDioxide"        , "CO_{2}"),
    ("SulfurHexafluoride"   , "SF_{6}"),
    ("Water"                , "H_{2}O"),
    ("HeavyWater"           , "D_{2}O"),
    ("Methanol"             , "CH_{3}OH"),
    ("Ethanol"              , "C_{2}H_{5}OH"),
    # ("Acetone"              , "C_{3}H_{6}O"),
]
# fmt: on
"""Subselection of the most common fluids to chose from.
It is a list of tuples, where each tuple contains 2 strings as follows ::

    tuple[0] (str): CoolProp name
    tuple[1] (str): Chemical name
"""

# ------------------------------------------------------------------------------
#   FluidProperties
# ------------------------------------------------------------------------------


class FluidProperties:
    """Evaluates thermodynamic fluid properties of the given fluid at the given
    temperature(s) in ``['C]`` and pressure(s) in ``[bar]``. The results are
    stored as properties to this class as ``numpy.ndarray`` arrays. Useful for
    working out dataseries.

    Example ::

        fluid = FluidProperties("Water", 20, 1)
        print(fluid.rho)  # [998.2065435]

        fluid = FluidProperties("Water", [20, 21, 22], 1)
        print(fluid.rho)  # [998.2065435  997.99487638 997.77288644]

    Args:
        coolprop_name (`str`):
            The CoolProp name of the fluid to evaluate.

        T_in_deg_C (`float` | `list[float]` | `numpy.ndarray[]`):
            Temperature ['C] to evaluate fluid properties at.

        P_in_bar (`float` | `list[float]` | `numpy.ndarray[]`):
            Pressure [bar] to evaluate fluid properties at.

    Properties:
        coolprop_name (`str`):
            CoolProp name of the fluid.

        formula (`str`):
            Chemical formula of the fluid.

        MW (`float`):
            Molecular weight [kg/mol].

        T (`numpy.ndarray[]`):
            Evaluated temperature [K].

        P (`numpy.ndarray[]`):
            Evaluated pressure [Pa].

        rho (`numpy.ndarray[]`):
            Density [kg/m^3].

        nu (`numpy.ndarray[]`):
            Kinematic viscosity [m^2/s].

        eta (`numpy.ndarray[]`):
            Dynamic/shear viscosity [kg/(m s)].

        alpha (`numpy.ndarray[]`):
            Thermal expansion coefficient [1/K].

        kappa (`numpy.ndarray[]`):
            Thermal diffusivity [m^2/s].

        lambda_ (`numpy.ndarray[]`):
            Thermal conductivity [W/(m K)].

        Cp (`numpy.ndarray[]`):
            Isobaric heat capacity [J/(kg K)].

        Cv (`numpy.ndarray[]`):
            Isochoric heat capacity [J/(kg K)].

        comp (`numpy.ndarray[]`):
            Isothermal compressibility [1/Pa].

        Pr (`numpy.ndarray[]`):
            Prandtl number.
    """

    def __init__(
        self,
        coolprop_name: str,
        T_in_deg_C: Union[float, list[float], npt.NDArray],
        P_in_bar: Union[float, list[float], npt.NDArray],
    ):
        # -------------------------
        #   Check input arguments
        # -------------------------

        if isinstance(T_in_deg_C, (float, int)):
            T_in_deg_C = np.array([T_in_deg_C], dtype=float)
        else:
            T_in_deg_C = np.asarray(T_in_deg_C, dtype=float)

        if isinstance(P_in_bar, (float, int)):
            P_in_bar = np.array([P_in_bar], dtype=float)
        else:
            P_in_bar = np.asarray(P_in_bar, dtype=float)

        if T_in_deg_C.ndim != 1 or P_in_bar.ndim != 1:
            raise ValueError(
                "Arguments `T_in_deg_C` and `P_in_bar` must be one-dimensional " "arrays."
            )

        if len(T_in_deg_C) > 1 and len(P_in_bar) == 1:
            P_in_bar = np.repeat(P_in_bar, len(T_in_deg_C))

        if len(P_in_bar) > 1 and len(T_in_deg_C) == 1:
            T_in_deg_C = np.repeat(T_in_deg_C, len(P_in_bar))

        if len(T_in_deg_C) != len(P_in_bar):
            raise ValueError("Arguments `T_in_deg_C` and `P_in_bar` have unequal lengths.")

        array_len = np.size(T_in_deg_C)
        nan_array = np.empty(array_len)
        nan_array[:] = np.nan

        # -----------
        #   Members
        # -----------

        self.coolprop_name: str = coolprop_name
        """CoolProp name of the fluid"""

        self.formula: str = ""
        """Chemical formula of the fluid"""

        self.MW: float = np.nan
        """Molecular weight [kg/mol]"""

        self.T: npt.NDArray[np.float64] = np.add(T_in_deg_C, ZERO_C)
        """Evaluated temperature [K]"""

        self.P: npt.NDArray[np.float64] = np.multiply(P_in_bar, 1e5)
        """Evaluated pressure [Pa]"""

        self.rho: npt.NDArray[np.float64] = np.copy(nan_array)
        """Density [kg/m^3]"""

        self.nu: npt.NDArray[np.float64] = np.copy(nan_array)
        """Kinematic viscosity [m^2/s]"""

        self.eta: npt.NDArray[np.float64] = np.copy(nan_array)
        """Dynamic/shear viscosity [kg/(m s)]"""

        self.alpha: npt.NDArray[np.float64] = np.copy(nan_array)
        """Thermal expansion coefficient [1/K]"""

        self.kappa: npt.NDArray[np.float64] = np.copy(nan_array)
        """Thermal diffusivity [m^2/s]"""

        self.lambda_: npt.NDArray[np.float64] = np.copy(nan_array)
        """Thermal conductivity [W/(m K)]"""

        self.Cp: npt.NDArray[np.float64] = np.copy(nan_array)
        """Isobaric heat capacity [J/(kg K)]"""

        self.Cv: npt.NDArray[np.float64] = np.copy(nan_array)
        """Isochoric heat capacity [J/(kg K)]"""

        self.comp: npt.NDArray[np.float64] = np.copy(nan_array)
        """Isothermal compressibility [1/Pa]"""

        self.Pr: npt.NDArray[np.float64] = np.copy(nan_array)
        """Prandtl number"""

        # ------------------------------
        #   Calculate fluid properties
        # ------------------------------

        found_match = False
        for item_ in FLUID_SELECTION:
            if item_[0] == coolprop_name:
                found_match = True
                self.formula = item_[1]
                break

        if not found_match:
            self.formula = CP.get_fluid_param_string(coolprop_name, "formula")
            self.formula = self.formula.replace("_{1}", "")

        # Molecular weight [kg/mol]
        self.MW = CP.PropsSI(coolprop_name, "M")

        # http://coolprop.org/coolprop/HighLevelAPI.html#table-of-string-inputs-to-propssi-function
        requested_quantities = [
            "DMASS",
            "VISCOSITY",
            "ISOBARIC_EXPANSION_COEFFICIENT",
            "CONDUCTIVITY",
            "CPMASS",
            "CVMASS",
            "ISOTHERMAL_COMPRESSIBILITY",
            "PRANDTL",
        ]

        for idx_, (T, P) in enumerate(zip(self.T, self.P)):
            values = np.zeros(len(requested_quantities))
            values[:] = np.nan

            for quantity_idx, quantity in enumerate(requested_quantities):
                try:
                    value = CP.PropsSI(quantity, "T", T, "P", P, coolprop_name)
                except ValueError as e:
                    value = np.nan
                    print(e)

                values[quantity_idx] = value

            # fmt: off
            (
                self.rho[idx_],     # Density                       [kg/m^3]
                self.eta[idx_],     # Dynamic viscosity             [kg/(m s)]
                self.alpha[idx_],   # Thermal expansion coefficient [1/K]
                self.lambda_[idx_], # Thermal conductivity          [W/(m K)]
                self.Cp[idx_],      # Isobaric heat capacity        [J/(kg K)]
                self.Cv[idx_],      # Isochoric heat capacity       [J/(kg K)]
                self.comp[idx_],    # Isothermal compressibility    [1/Pa]
                self.Pr[idx_],      # Prandtl number                [-]
            ) = values
            # fmt: on

        # Derived: Kinematic viscosity [m^2/s]
        self.nu = self.eta / self.rho
        # Derived: Thermal diffusivity [m^2/s]
        self.kappa = self.lambda_ / self.rho / self.Cp



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Module containing class to convert a string to asdl class type and vica-versa."""
import esdl

from omotes_simulator_core.adapter.transforms.transform_utils import reverse_dict
from typing import List


class StringEsdlAssetMapper:
    """Mapper class to convert strings to an esdl class type and vica-versa.

    Please note that the str_to_type_dict needs to have unique keys and values.
    Also note that fo the esdl class only the derived classes should be listed.
    Otherwise asset might be selected twice, since a HeatingDemand is also and Consumer.
    """

    type_to_str_dict = {
        esdl.Asset: "asset",
        esdl.GenericProducer: "producer",
        esdl.GenericConsumer: "consumer",
        esdl.HeatingDemand: "consumer",
        esdl.GeothermalSource: "geothermal",
        esdl.Conversion: "conversion",
        esdl.Pipe: "pipe",
        esdl.Transport: "transport",
        esdl.Joint: "joint",
        esdl.ATES: "storage",
    }

    str_to_type_dict = reverse_dict(original_dict=type_to_str_dict)

    def to_esdl(self, entity: str) -> List[type]:
        """Method to convert a string to esdl class type.

        :param str entity: String to be converted to an esdl class type object.
        :return: ESDL class described in the string.
        """
        if entity not in self.str_to_type_dict:
            raise NotImplementedError(entity + " not implemented in StringESDLAssetMapper class")
        return self.str_to_type_dict[entity]

    def to_entity(self, entity: type) -> str:
        """Method to convert esdl type class to a string.

        The dict values are converted to a list to find the index in the list.
        Then the list of keys is created and the value at the index found is returned.
        :param type entity: ESDl object clas to be converted to a string
        :return: str belonging to the entity.
        """
        if entity not in self.type_to_str_dict:
            raise NotImplementedError(
                str(entity) + " not implemented in StringESDLAssetMapper class"
            )
        return self.type_to_str_dict[entity]



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Plot result on the map for debugging."""
import esdl
import plotly.express as px
import geopandas as gpd
from pandas import DataFrame
import numpy as np
from shapely.geometry import LineString
from numpy.typing import NDArray
from omotes_simulator_core.entities.esdl_object import EsdlObject
from omotes_simulator_core.entities.assets.asset_defaults import (
    PROPERTY_MASSFLOW,
    PROPERTY_PRESSURE,
    PROPERTY_TEMPERATURE,
)


class Plotting:
    """Plotting plot the results of asset (currently pipe) for inspection."""

    def __init__(self, esdl_object: EsdlObject) -> None:
        """Initialize plotting instance.

        :param EsdlObject esdl_object: esdl object of the network
        """
        self.esdl_object = esdl_object

    def simulation_output(self, result: DataFrame) -> None:
        """Function to plot the simulation results to the map.

        :param DataFrame result: simulation result in dataframe format
        """
        geo_df = self.prepare_data(result)
        Plotting.plot_line_mapbox(geo_df)

    def prepare_data(self, result: DataFrame) -> gpd.GeoDataFrame:
        """Function to prepare geopandas dataframe."""
        geo_df = gpd.GeoDataFrame(
            {},
            columns=[
                "asset_id",
                "asset_name",
                "geometry",
                "mass_flow",
                "pressure_in",
                "pressure_out",
                "temperature_in",
                "temperature_out",
            ],
            crs="EPSG:4326",
        )

        assets = self.esdl_object.get_all_assets_of_type("pipe")

        for asset_object in assets:
            asset = asset_object.esdl_asset
            if isinstance(asset.geometry, esdl.esdl.Line):
                coor_pipe = []
                for point in asset.geometry.point:
                    coor_pipe.append([point.lon, point.lat])

                last_data_index = len(result) - 1
                data = {
                    "asset_id": asset.id,
                    "asset_name": asset.name,
                    "geometry": LineString(coor_pipe),
                    "mass_flow": result[(asset_object.get_port_ids()[0], PROPERTY_MASSFLOW)][
                        last_data_index
                    ],
                    "pressure_in": result[(asset_object.get_port_ids()[0], PROPERTY_PRESSURE)][
                        last_data_index
                    ],
                    "pressure_out": result[(asset_object.get_port_ids()[1], PROPERTY_PRESSURE)][
                        last_data_index
                    ],
                    "temperature_in": result[
                        (asset_object.get_port_ids()[0], PROPERTY_TEMPERATURE)
                    ][last_data_index],
                    "temperature_out": result[
                        (asset_object.get_port_ids()[1], PROPERTY_TEMPERATURE)
                    ][last_data_index],
                }

                geo_df.loc[len(geo_df)] = data

        return geo_df

    @staticmethod
    def plot_line_mapbox(geo_df: gpd.GeoDataFrame, output_filename: str = "map.html") -> None:
        """Function to plot using plotly express line mapbox."""
        lats: NDArray = np.array(None)
        lons: NDArray = np.array(None)
        names: NDArray = np.array(None)
        for _, row in geo_df.iterrows():
            if isinstance(row["geometry"], LineString):
                linestrings = [row["geometry"]]
            else:
                continue
            for linestring in linestrings:
                x, y = linestring.xy
                lats = np.append(lats, y)
                lons = np.append(lons, x)

                text = (
                    row["asset_name"]
                    + ", mass_flow: "
                    + "{:.2f}".format(row["mass_flow"])
                    + " kg/s"
                    + ", Pin: "
                    + "{:.2f}".format(row["pressure_in"] / 1e5)
                    + " bar"
                    + ", Pout: "
                    + "{:.2f}".format(row["pressure_out"] / 1e5)
                    + " bar"
                    + ", Tin: "
                    + "{:.2f}".format(row["temperature_in"] - 273)
                    + " C"
                    + ", Tout: "
                    + "{:.2f}".format(row["temperature_out"] - 273)
                    + " C"
                )

                names = np.append(names, [text] * len(y))
                lats = np.append(lats, np.array(None))
                lons = np.append(lons, np.array(None))
                names = np.append(names, np.array(None))

        fig = px.line_mapbox(
            lat=lats, lon=lons, hover_name=names, zoom=14, mapbox_style="carto-positron"
        )
        fig.write_html(output_filename)



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module containing the Esdl to asset mapper class."""
from typing import Any, Type

import esdl
import numpy as np

from omotes_simulator_core.entities.assets.asset_abstract import AssetAbstract
from omotes_simulator_core.entities.assets.demand_cluster import DemandCluster
from omotes_simulator_core.entities.assets.esdl_asset_object import EsdlAssetObject
from omotes_simulator_core.entities.assets.pipe import Pipe
from omotes_simulator_core.entities.assets.production_cluster import ProductionCluster
from omotes_simulator_core.entities.assets.ates_cluster import AtesCluster
from omotes_simulator_core.entities.assets.heat_pump import HeatPump

from omotes_simulator_core.entities.assets.controller.controller_producer import ControllerProducer
from omotes_simulator_core.entities.assets.controller.controller_consumer import ControllerConsumer
from omotes_simulator_core.entities.assets.controller.controller_storage import ControllerStorage
from omotes_simulator_core.simulation.mappers.mappers import EsdlMapperAbstract, Entity

CONVERSION_DICT: dict[esdl.EnergyAsset, Type[AssetAbstract]] = {
    esdl.Producer: ProductionCluster,
    esdl.GenericProducer: ProductionCluster,
    esdl.Consumer: DemandCluster,
    esdl.GenericConsumer: DemandCluster,
    esdl.HeatingDemand: DemandCluster,
    esdl.Pipe: Pipe,
    esdl.ATES: AtesCluster,
    esdl.HeatPump: HeatPump,
}


class EsdlAssetMapper:
    """Creates entity Asset objects based on a PyESDL EnergySystem assets."""

    @staticmethod
    def to_esdl(entity: AssetAbstract) -> Any:
        """Maps entity object to PyEsdl objects."""
        raise NotImplementedError("EsdlAssetMapper.to_esdl()")

    @staticmethod
    def to_entity(model: EsdlAssetObject) -> AssetAbstract:
        """Method to map an esdl asset to an asset entity class.

        :param EsdlAssetObject model: Object to be converted to an asset entity.

        :return: Entity object of type AssetAbstract.
        """
        if not type(model.esdl_asset) in CONVERSION_DICT:
            raise NotImplementedError(str(model.esdl_asset) + " not implemented in conversion")
        return CONVERSION_DICT[type(model.esdl_asset)](
            model.esdl_asset.name, model.esdl_asset.id, model.get_port_ids()
        )


class EsdlAssetControllerProducerMapper(EsdlMapperAbstract):
    """Class to map an esdl asset to a producer entity class."""

    def to_esdl(self, entity: Entity) -> EsdlAssetObject:
        """Map an Entity to a EsdlAsset."""
        raise NotImplementedError("EsdlAssetControllerProducerMapper.to_esdl()")

    def to_entity(self, esdl_asset: EsdlAssetObject) -> ControllerProducer:
        """Method to map an esdl asset to a producer entity class.

        :param EsdlAssetObject model: Object to be converted to an asset entity.

        :return: Entity object.
        """
        result = esdl_asset.get_property(esdl_property_name="power", default_value=0)
        if result[1]:
            power = result[0]
        else:
            raise ValueError("No power found for asset: " + esdl_asset.esdl_asset.name)
        marginal_costs = esdl_asset.get_marginal_costs()
        temperature_supply = esdl_asset.get_supply_temperature("Out")
        temperature_return = esdl_asset.get_return_temperature("In")
        contr_producer = ControllerProducer(
            name=esdl_asset.esdl_asset.name,
            identifier=esdl_asset.esdl_asset.id,
            temperature_supply=temperature_supply,
            temperature_return=temperature_return,
            power=power,
            marginal_costs=marginal_costs,
        )
        return contr_producer


class EsdlAssetControllerConsumerMapper(EsdlMapperAbstract):
    """Class to map an esdl asset to a consumer entity class."""

    def to_esdl(self, entity: Entity) -> EsdlAssetObject:
        """Map an Entity to a EsdlAsset."""
        raise NotImplementedError("EsdlAssetControllerProducerMapper.to_esdl()")

    def to_entity(self, esdl_asset: EsdlAssetObject) -> ControllerConsumer:
        """Method to map an esdl asset to a consumer entity class.

        :param EsdlAssetObject model: Object to be converted to an asset entity.

        :return: Entity object.
        """
        result = esdl_asset.get_property(esdl_property_name="power", default_value=np.inf)
        power = np.inf
        if result[1]:
            power = result[0]
        if power == 0:
            power = np.inf

        # It looks like they are switch, but this is because of the definition used in ESDL,
        # which is different as what we use.
        temperature_supply = esdl_asset.get_return_temperature("Out")
        temperature_return = esdl_asset.get_supply_temperature("In")
        profile = esdl_asset.get_profile()
        contr_consumer = ControllerConsumer(
            name=esdl_asset.esdl_asset.name,
            identifier=esdl_asset.esdl_asset.id,
            temperature_supply=temperature_supply,
            temperature_return=temperature_return,
            max_power=power,
            profile=profile,
        )
        return contr_consumer


class EsdlAssetControllerStorageMapper(EsdlMapperAbstract):
    """Class to map an esdl asset to a storage entity class."""

    def to_esdl(self, entity: Entity) -> EsdlAssetObject:
        """Map an Entity to a EsdlAsset."""
        raise NotImplementedError("EsdlAssetControllerStorageMapper.to_esdl()")

    def to_entity(self, esdl_asset: EsdlAssetObject) -> ControllerStorage:
        """Method to map an esdl asset to a storage entity class.

        :param EsdlAssetObject model: Object to be converted to an asset entity.

        :return: Entity object.
        """
        result = esdl_asset.get_property(
            esdl_property_name="maxDischargeRate", default_value=np.inf
        )
        discharge_power = np.inf
        if result[1]:
            discharge_power = result[0]

        result = esdl_asset.get_property(esdl_property_name="maxChargeRate", default_value=np.inf)
        charge_power = np.inf
        if result[1]:
            charge_power = result[0]

        temperature_supply = esdl_asset.get_supply_temperature("In")
        temperature_return = esdl_asset.get_return_temperature("Out")
        profile = esdl_asset.get_profile()
        contr_storage = ControllerStorage(
            name=esdl_asset.esdl_asset.name,
            identifier=esdl_asset.esdl_asset.id,
            temperature_supply=temperature_supply,
            temperature_return=temperature_return,
            max_charge_power=charge_power,
            max_discharge_power=discharge_power,
            profile=profile,
        )
        return contr_storage



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Module containing the production asset class."""
import numpy as np

from omotes_simulator_core.solver.matrix.equation_object import EquationObject
from omotes_simulator_core.solver.network.assets.fall_type import FallType


class ProductionAsset(FallType):
    """
    A class to represent a production asset in a network.

    This class inherits from the FallType class and implements the methods to generate
    the equations for the production asset.

    Attributes
    ----------
    connected_nodes : list[int]
        A list of integers that store the node indices that are connected to the asset.
    number_of_connection_point : int
        The number of connection points for the asset.
    pre_scribe_mass_flow : bool
        A boolean flag that indicates whether the mass flow rate or the pressure is prescribed
        at the connection points.

    Methods
    -------
    get_equations() -> list[EquationObject]
        Returns a list of EquationObjects that represent the equations for the asset.
    get_pre_scribe_mass_flow_or_pressure_equations(connection_point: int) -> EquationObject
        Returns an EquationObject that represents the prescribed mass flow rate or pressure
        equation for the asset at the given connection point.
    """

    mass_flow_rate_set_point: float
    """The mass flow rate set point for the asset."""

    set_pressure: float
    """The pressure set point for the asset."""

    pre_scribe_mass_flow: bool
    """A boolean flag that indicates whether the mass flow rate or the pressure is prescribed
    at the connection points."""

    def __init__(
        self,
        name: str,
        _id: str,
        supply_temperature: float = 293.15,
        heat_supplied: float = 0.0,
        loss_coefficient: float = 1.0,
        pre_scribe_mass_flow: bool = True,
        mass_flow_rate_set_point: float = 10.0,
        set_pressure: float = 10000.0,
    ):
        """
        Initializes the ProductionAsset object with the given parameters.

        Parameters
        ----------
        name : str The name of the asset.
        _id : str The unique identifier of the asset.
        number_of_unknowns : int, optional
            The number of unknown variables for the asset. The default is 6, which corresponds to
            the mass flow rate, pressure, and temperature at each connection point.
        number_con_points : int, optional
            The number of connection points for the asset. The default is 2, which corresponds to
            the inlet and outlet.
        """
        self.pre_scribe_mass_flow = pre_scribe_mass_flow
        self.mass_flow_rate_set_point = mass_flow_rate_set_point
        self.set_pressure = set_pressure

        super().__init__(
            name=name,
            _id=_id,
            supply_temperature=supply_temperature,
            heat_supplied=heat_supplied,
            loss_coefficient=loss_coefficient,
        )

    def get_equations(self) -> list[EquationObject]:
        """Returns a list of EquationObjects that represent the equations for the asset.

        The equations are:
        - Pressure balance at each connection point
        - Thermal balance at each connection point
        - Prescribed mass flow rate or pressure at each connection point
        :return: list[EquationObject]
            A list of EquationObjects that contain the indices, coefficients, and right-hand side
            values of the equations.
        """
        equations = [
            super().get_press_to_node_equation(0),
            super().get_press_to_node_equation(1),
            self.get_thermal_equations(0),
            self.get_thermal_equations(1),
            self.get_pre_scribe_mass_flow_or_pressure_equations(0),
            self.get_pre_scribe_mass_flow_or_pressure_equations(1),
        ]
        return equations

    def get_pre_scribe_mass_flow_or_pressure_equations(
        self, connection_point: int
    ) -> EquationObject:
        """Returns an EquationObject for a pre describe equation.

        The returned equation object represents the prescribed mass flow rate or pressure
        equation for the asset at the given connection point.

        The equation is:

        - If pre_scribe_mass_flow is True, then Mass flow rate at connection point = Mass flow rate
        property
        - If pre_scribe_mass_flow is False, then Pressure at connection point = Set pressure
        property
        :param int connection_point: The connection point for which to get the equation
        :return: EquationObject
            An EquationObject that contains the indices, coefficients, and right-hand side
            value of the equation.
        """
        # Raise IndexError if the connection point is not available
        if connection_point >= self.number_of_connection_point:
            raise IndexError("The connection point is not available.")

        # Create equation object
        equation_object = EquationObject()
        if self.pre_scribe_mass_flow:
            equation_object.indices = np.array(
                [
                    self.get_index_matrix(
                        property_name="mass_flow_rate",
                        connection_point=connection_point,
                        use_relative_indexing=False,
                    )
                ]
            )
            equation_object.coefficients = np.array([-1.0 + 2 * connection_point])
            equation_object.rhs = self.mass_flow_rate_set_point
        else:
            equation_object.indices = np.array(
                [
                    self.get_index_matrix(
                        property_name="pressure",
                        connection_point=connection_point,
                        use_relative_indexing=False,
                    )
                ]
            )
            equation_object.coefficients = np.array([1.0])
            if connection_point == 0:
                equation_object.rhs = 0.5 * self.set_pressure
            else:
                equation_object.rhs = self.set_pressure
        return equation_object

    def get_thermal_equations(self, connection_point: int) -> EquationObject:
        """Gets a thermal equation for a connection point of the asset.

        :param connection_point: The index of the connection point to get the equation for.
        :type connection_point: int
        :return: An equation object representing the thermal equation.
        :rtype: EquationObject
        """
        if (
            self.prev_sol[
                self.get_index_matrix(
                    property_name="mass_flow_rate",
                    connection_point=connection_point,
                    use_relative_indexing=True,
                )
            ]
            > 0
        ):
            return self.get_prescribe_temp_equation(connection_point)
        else:
            return self.get_internal_energy_to_node_equation(connection_point)



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""HeatNetwork entity class."""
import datetime
from typing import Callable, List, Tuple

import pandas as pd

from omotes_simulator_core.entities.assets.asset_abstract import AssetAbstract
from omotes_simulator_core.entities.assets.junction import Junction
from omotes_simulator_core.solver.network.network import Network
from omotes_simulator_core.solver.solver import Solver


class HeatNetwork:
    """Class to store information on the heat network."""

    def __init__(
        self, conversion_factory: Callable[[Network], Tuple[List[AssetAbstract], List[Junction]]]
    ) -> None:
        """Constructor of heat network class.

        :param conversion_factory: method to convert the esdl network to lists of assets&junctions
        and returns list of both
        """
        self.network = Network()
        self.assets, self.junctions = conversion_factory(self.network)
        self.solver = Solver(self.network)

    def run_time_step(self, time: datetime.datetime, controller_input: dict) -> None:
        """Method to simulate a time step.

        It first sets the controller input to the assets and then simulates the time step.

        :param float time: Timestep for which to simulate the model
        :param dict controller_input: Dict specifying the heat demand for the different assets.
        :return: None
        """
        for py_asset in self.assets:
            if py_asset.asset_id in controller_input:
                py_asset.set_setpoints(controller_input[py_asset.asset_id])
        self.solver.solve()

    def plot_network(self) -> None:
        """Method to plot the network.

        plots the network in a simple plot all junctions are translated to a circle.
        :return:
        """

    def store_output(self) -> None:
        """Method to store the output data.

        This method takes the data from the assets and stores it into our own
        dataframe. This is needed since we have the possibility to redo a timestep when results are
        not converged for the input of the controller.
        :return: None
        """
        for py_asset in self.assets:
            py_asset.write_standard_output()
            py_asset.write_to_output()

    def gather_output(self) -> pd.DataFrame:
        """Method to gather output of all assets and return it as a dict.

        :return: Dict with for all asset a dict with all outputs.
        """
        # Todo what do we do with the junction output do we need it?
        result = pd.DataFrame()
        for py_asset in self.assets:
            result = pd.concat([result, py_asset.get_timeseries()], axis=1)
        return result



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This file contains the class for the core quantities of the matrix."""
import dataclasses


@dataclasses.dataclass
class IndexCoreQuantity:
    """Enum for the index of the matrix.

    This is used to store the order of the core quants in the matrix.
    The number of core quantities is the maximum number of core quantities used.
    for the indices these should be in increasing order. This is not checked.
    """

    number_core_quantities = 3
    mass_flow_rate = 0
    pressure = 1
    internal_energy = 2

    def get_index_property(self, property_name: str, connection_point: int) -> int:
        """Method to get the property of the index."""
        return self.get_index(property_name) + connection_point * self.number_core_quantities

    def get_index(self, property_name: str) -> int:
        """Method to get the index of the property."""
        return int(getattr(self, property_name))


index_core_quantity = IndexCoreQuantity()



#  Copyright (c) 2024. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Module containing the classes for the controller."""

from omotes_simulator_core.entities.assets.controller.asset_controller_abstract import (
    AssetControllerAbstract,
)


class ControllerProducer(AssetControllerAbstract):
    """Class to store the source for the controller."""

    def __init__(
        self,
        name: str,
        identifier: str,
        temperature_supply: float,
        temperature_return: float,
        power: float,
        marginal_costs: float,
        priority: int = 1,
    ):
        """Constructor for the source.

        :param str name: Name of the source.
        :param str identifier: Unique identifier of the source.
        :param float temperature_supply: Supply temperature of the source.
        :param float temperature_return: Return temperature of the source.
        :param float power: Power of the source.
        :param float marginal_costs: Marginal costs of the source.
        :param int priority: Priority of the source.
        """
        super().__init__(name, identifier)
        self.temperature_return: float = temperature_return
        self.temperature_supply: float = temperature_supply
        self.power: float = power
        self.marginal_costs: float = marginal_costs
        self.priority: int = priority



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""This module contains the fluid properties class."""

import numpy as np
from omotes_simulator_core.solver.utils.fluidprop import FluidProperties


class Interpolation:
    """Class to enable interpolation in a set of data points."""

    def __init__(self, x: list[float], y: list[float], order: int = 5, bounds: float = 0.1):
        """Constructor of the interpolation class.

        It stores the x and y data and determine the coefficients of a polynomial of the given
        order. This polynomial is used to interpolate the data.

        :param x: The x values of the data points.
        :param y: The y values of the data points.
        :param order: The order of the polynomial to use for the interpolation.
        :param bounds: The bounds on the x values used to check if the value is within bounds.
        """
        self.x = x
        self.y = y
        self.order = order
        self.bounds = bounds
        self.coefficients = np.polyfit(x, y, order)
        self._check_interpolation()

    def _check_interpolation(self) -> None:
        """Check if the interpolation gives acceptable results.

        For every data point the difference with the interpolated value is calculated.
        If this difference is more than 2% and value error is raised.
        """
        error = [(y - self(x)) / y for x, y in zip(self.x, self.y)]
        if any(abs(e) > 0.02 and abs(e) != np.inf for e in error):
            raise ValueError("Interpolation error: error is more then 2%.")

    def __call__(self, value: float) -> float:
        """Returns the interpolated value at a given point.

        The calculate the value at the given point, the coefficients of the polynomial are used.
        A backwards loop is used, since the first value in the list is the one which is multiplied
        with the highest order. In this way we can step by step multiply the temp value with the
        value at the given point.

        :param value: The value to interpolate.
        :return: The interpolated value at the given point.
        """
        self._check_bounds(value=value)
        result = 0
        temp_value = 1.0
        for i in range(self.order, -1, -1):
            result += self.coefficients[i] * temp_value
            temp_value *= value
        return result

    def _check_bounds(self, value: float) -> None:
        """Check if the value is within the bounds of the data.

        The bounds check is within the set bounds default 10%, since we are fitting a curve
        on the data points. The result is that the curve can be a bit wider than the data points.
        """
        if value < self.x[0] * (1.0 - self.bounds) or value > self.x[-1] * (1.0 + self.bounds):
            raise ValueError("Value is out of bounds.")


class OmotesFluidProperties:
    """Class to represent the fluid properties.

    This class contains methods to get the internal energy, temperature, density,
    and viscosity of the fluid. THe data is loaded from a csv file in the same folder as this file.
    """

    T: list[float]
    """A list of floats that store the temperature of the fluid [K] per temperature."""

    cp: list[float]
    """A list of floats that store the heat capacity of the fluid [J/kg/K] per temperature."""

    rho: list[float]
    """A list of floats that store the density of the fluid [kg/m^3] per temperature."""

    visc: list[float]
    """A list of floats that store the viscosity of the fluid [Pa.s] per temperature."""

    therm_cond: list[float]
    """A list of floats that store the thermal conductivity of the fluid [W/m/K] per temperature."""

    IE: list[float]
    """A list of floats that store the internal energy of the fluid [J/kg] per temperature."""

    def __init__(
        self, p_ref: float = 20, fluid: str = "Water", T_min: int = 0, T_max: int = 150
    ) -> None:
        """Constructor of the fluid properties class.

        Initializes the class properties. Loads a list of the fluid property as function of the
        temperature using the fluidprop library. These are then used interpolation class objects.
        :param p_ref: The reference pressure of the fluid.
        :param fluid: The fluid to use for the fluid properties.
        :param T_min: The minimum temperature in Celsius to load the fluid properties for.
        :param T_max: The maximum temperature in Celsius to load the fluid properties for.
        """
        self.T = []
        self.cp = []
        self.rho = []
        self.visc = []
        self.therm_cond = []

        for t in range(T_min, T_max):
            self.T.append(t + 273.15)
            fluidprops = FluidProperties(fluid, t, p_ref)
            self.cp.append(float(fluidprops.Cp[0]))
            self.rho.append(float(fluidprops.rho[0]))
            self.visc.append(float(fluidprops.nu[0]))
            self.therm_cond.append(float(fluidprops.lambda_[0]))

        self.IE = [0.0]
        for i in range(1, len(self.T)):
            self.IE.append(
                self.IE[-1] + (self.cp[i - 1] + self.cp[i]) / 2 * (self.T[i] - self.T[i - 1])
            )

        # Create interpolation objects for the fluid properties
        self.ie_func = Interpolation(self.T, self.IE, 5)
        self.temp_func = Interpolation(self.IE, self.T, 5)
        self.density_func = Interpolation(self.T, self.rho, 5)
        self.visc_func = Interpolation(self.T, self.visc, 6)
        self.heat_cap_func = Interpolation(self.T, self.cp, 5)
        self.therm_cond_func = Interpolation(self.T, self.therm_cond, 5)

    def get_ie(self, t: float) -> float:
        """Returns the internal energy of the fluid at a given temperature.

        :param t: The temperature of the fluid.
        :return: The internal energy of the fluid at the given temperature.
        """
        return self.ie_func(t)

    def get_t(self, ie: float) -> float:
        """Returns the temperature of the fluid at a given internal energy.

        :param ie: The internal energy of the fluid.
        :return: The temperature of the fluid at the given internal energy.
        """
        return self.temp_func(ie)

    def get_density(self, t: float) -> float:
        """Returns the density of the fluid at a given temperature.

        :param t: The temperature of the fluid.
        :return: The density of the fluid at the given temperature.
        """
        return self.density_func(t)

    def get_viscosity(self, t: float) -> float:
        """Returns the viscosity of the fluid at a given temperature.

        :param t: The temperature of the fluid.
        :return: The viscosity of the fluid at the given temperature.
        """
        return self.visc_func(t)

    def get_heat_capacity(self, t: float) -> float:
        """Returns the heat capacity of the fluid at a given temperature.

        :param t: The temperature of the fluid.
        :return: The capacity of the fluid at the given temperature.
        """
        return self.heat_cap_func(t)

    def get_thermal_conductivity(self, t: float) -> float:
        """Returns the thermal conductivity of the fluid at a given temperature.

        :param t: The temperature of the fluid.
        :return: The thermal conductivity of the fluid at the given temperature.
        """
        return self.therm_cond_func(t)


fluid_props = OmotesFluidProperties()



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Simulates an heat network for the specified duration."""
from typing import Callable
from pandas import DataFrame
from omotes_simulator_core.entities.heat_network import HeatNetwork
from omotes_simulator_core.entities.network_controller import NetworkController
from omotes_simulator_core.entities.simulation_configuration import SimulationConfiguration
from datetime import timedelta, timezone

import logging

logger = logging.getLogger(__name__)

MAX_NUMBER_MESSAGES = 15


class NetworkSimulation:
    """NetworkSimulation connects the controller and HeatNetwork (incl. assets)."""

    def __init__(self, network: HeatNetwork, controller: NetworkController):
        """Instantiate the NetworkSimulation object."""
        self.network = network
        self.controller = controller

    def run(
        self,
        config: SimulationConfiguration,
        progress_calback: Callable[[float, str], None],
        max_number_messages: int = MAX_NUMBER_MESSAGES,
    ) -> None:
        """Run the simulation.

        :param SimulationConfiguration config: Configuration parameters for simulation.
        :param Callable[[float, str], None] progress_calback: Callback function to report progress.
        :param int max_number_messages: Maximum number of messages to report progress.
        """
        # time loop
        number_of_time_steps = int((config.stop - config.start).total_seconds() / config.timestep)
        logger.info("Number of time steps: " + str(number_of_time_steps))
        progress_interval = max(round(number_of_time_steps / max_number_messages), 1)
        for time_step in range(number_of_time_steps):
            not_converged = True
            time = (config.start + timedelta(seconds=time_step * config.timestep)).replace(
                tzinfo=timezone.utc
            )
            controller_input = self.controller.update_setpoints(time)
            logger.debug("Simulating for timestep " + str(time))

            while not_converged:
                not_converged = False  # for the moment we do not check on convergence,
                # to get stuff running. Also need to add break after 10 iteration.
                self.network.run_time_step(time, controller_input)
            self.network.store_output()

            if (time_step % progress_interval) == 0:
                progress_calback((float(time_step) / float(number_of_time_steps)), "calculating")

    def gather_output(self) -> DataFrame:
        """Gathers all output and return a dict with this output.

        :return: DataFrame with all the results for the simulation
        """
        result = self.network.gather_output()
        return result



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""File containing utility functions for the transforms."""
from typing import Dict, List
from enum import Enum
from dataclasses import dataclass


class PortType(Enum):
    """Enum to define the type of port."""

    IN = 1
    OUT = 2


@dataclass
class Port:
    """Dataclass to hold the port information."""

    port_id: str
    port_name: str
    port_type: PortType


def sort_ports(connected_ports: list[Port]) -> list[str]:
    """Sort the ports of the asset based on the port type.

    The sort order is Inport, Outport, Starting with the first inport, then the first outport
    and so on.

    :param connected_ports: List of tuples with the port id, name and the port type.
    :return: List of port ids sorted by port type.
    """
    in_ports = [port for port in connected_ports if port.port_type == PortType.IN]
    out_ports = [port for port in connected_ports if port.port_type == PortType.OUT]
    if len(in_ports) != len(out_ports):
        raise ValueError("The number of in ports and out ports are not equal")
    sorted_port_list = []
    for in_port, out_port in zip(in_ports, out_ports):
        sorted_port_list.append(in_port)
        sorted_port_list.append(out_port)
    if len(sorted_port_list) == 4:
        sorted_port_list = order_prim_sec_ports(sorted_port_list)
    return [port.port_id for port in sorted_port_list]


def order_prim_sec_ports(connected_ports: list[Port]) -> list[Port]:
    """Order the primary and secondary ports. in correct order.

    The correct order is first primary port, then secondary port.
    This can only be done by checking the port names.
    In primary port prim is in the name for secondary port sec is in the name.

    :param connected_ports: List of connected ports to be sorted.
    :return: List of connected ports sorted by primary and secondary ports.
    """
    primary_ports = [port for port in connected_ports if "Prim" in port.port_name]
    if len(primary_ports) != 2:
        raise ValueError("The number of ports with prim in the name is not equal to 2")
    secondary_ports = [port for port in connected_ports if "Sec" in port.port_name]
    if len(secondary_ports) != 2:
        raise ValueError("The number of ports with sec in the name is not equal to 2")
    return primary_ports + secondary_ports


def reverse_dict(original_dict: dict) -> Dict[str, List[type]]:
    """Method to reverse a dict.

    Creates a dict with set(values) as keys and the keys of the original dict as list values.

    :param dict original_dict: Dict to be reversed {keys:values}.
    :return: Reversed dict with {set(values):list[keys]}.
    """
    new_dict = {}
    for key, value in original_dict.items():
        if value not in new_dict:
            new_dict[value] = [key]
        else:
            new_dict[value].append(key)
    return new_dict



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Utility functions for assets."""
from enum import IntEnum

import numpy as np
from omotes_simulator_core.entities.assets.esdl_asset_object import EsdlAssetObject
from typing import List, Tuple
from omotes_simulator_core.solver.utils.fluid_properties import fluid_props


def heat_demand_and_temperature_to_mass_flow(
    thermal_demand: float, temperature_supply: float, temperature_return: float
) -> float:
    """Calculate the mass flow rate that is required to meet the thermal demand.

    :param float thermal_demand: The thermal demand of the asset. The thermal demand should be
        supplied in Watts.
    :param float temperature_supply: The temperature that the asset delivers to the "to_junction".
        The temperature should be supplied in Kelvin. The supply temperature is used to calculate
        the specific heat capacity of the fluid.
    :param float temperature_return: The temperature that the asset receives from the
        "from_junction". The temperature should be supplied in Kelvin.
    """
    heat_capacity = fluid_props.get_heat_capacity((temperature_return + temperature_supply) / 2)
    return thermal_demand / ((temperature_supply - temperature_return) * float(heat_capacity))


def mass_flow_and_temperature_to_heat_demand(
    temperature_supply: float,
    temperature_return: float,
    mass_flow: float,
) -> float:
    """Calculate the thermal demand that is met by the mass flow rate.

    :param float temperature_supply: The temperature that the asset delivers to the "to_junction".
        The temperature should be supplied in Kelvin. The supply temperature is used to calculate
        the specific heat capacity of the fluid.
    :param float temperature_return: The temperature that the asset receives from the
        "from_junction". The temperature should be supplied in Kelvin.
    :param float mass_flow: The mass flow rate that is used to meet the thermal demand. The mass
        flow rate should be supplied in kg/s.
    """
    internal_energy1 = fluid_props.get_ie(temperature_supply)
    internal_energy2 = fluid_props.get_ie(temperature_return)
    return mass_flow * (internal_energy1 - internal_energy2)


def get_thermal_conductivity_table(esdl_asset: EsdlAssetObject) -> Tuple[List[float], List[float]]:
    """Retrieve the thermal conductivity table of the asset.

    :param EsdlAssetObject esdl_asset: The asset of which the heat transfer table should be
        retrieved.
    :return: The diameters and thermal conductivities of the asset in a tuple with the diameters
        as the first element and the thermal conductivities as the second element. Each element
        represents a layer of the asset.
    """
    # create heat transfer table
    layer_thicknesses = []
    diameters = []
    heat_coef = []
    if esdl_asset.esdl_asset.material is not None:
        if hasattr(esdl_asset.esdl_asset.material, "component"):
            material_object = esdl_asset.esdl_asset.material.component
        elif hasattr(esdl_asset.esdl_asset.material, "reference"):
            material_object = esdl_asset.esdl_asset.material.reference.component
        else:
            raise NotImplementedError("Unknown material type or reference.")
        # Append initial internal diameter
        diameters.append(esdl_asset.esdl_asset.innerDiameter)
        # Loop over material object
        for layer in material_object:
            layer_thicknesses.append(layer.layerWidth * 2)
            diameters.append(sum(layer_thicknesses) + esdl_asset.esdl_asset.innerDiameter)
            heat_coef.append(layer.matter.thermalConductivity)
    return diameters, heat_coef


def calculate_inverse_heat_transfer_coefficient(
    inner_diameter: np.ndarray, outer_diameter: np.ndarray, thermal_conductivity: np.ndarray
) -> np.ndarray:
    """Calculate the inverse heat transfer coefficient of a pipe.

    :param inner_diameter: Inner diameter of the pipe in m
    :param outer_diameter: Outer diameter of the pipe in m
    :param thermal_conductivity: Thermal conductivity of the pipe material in W/(m K)
    :return: Inverse heat transfer coefficient in W/(m^2 K)
    """
    return np.array(
        (inner_diameter * np.log(outer_diameter / inner_diameter)) / (2.0 * thermal_conductivity)
    )


class Port(IntEnum):
    """Simple enumeration class to set if it is In (from) or out(to) port of asset."""

    In = 0
    Out = 1



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Module containing pipe class."""
from typing import Dict, List

import numpy as np

from omotes_simulator_core.entities.assets.asset_abstract import AssetAbstract
from omotes_simulator_core.entities.assets.asset_defaults import PIPE_DEFAULTS
from omotes_simulator_core.entities.assets.esdl_asset_object import EsdlAssetObject
from omotes_simulator_core.entities.assets.utils import (
    calculate_inverse_heat_transfer_coefficient,
    get_thermal_conductivity_table,
)
from omotes_simulator_core.solver.network.assets.solver_pipe import SolverPipe


class Pipe(AssetAbstract):
    """A class representing a pipe in a heat network."""

    _minor_loss_coefficient: float
    """The minor loss coefficient of the pipe [-]."""
    _external_temperature: float
    """The external temperature surrounding the pipe [K]."""
    _qheat_external: float
    """The external heat flow into the pipe [W]."""
    length: float
    """The length of the pipe [m]."""
    diameter: float
    """The diameter of the pipe [m]."""
    roughness: float
    """The wall roughness of the pipe [m]."""
    alpha_value: float
    """The alpha value of the pipe [W/(m2 K)]."""
    output: List[Dict[str, float]]
    """The output list of the pipe with a dictionaries for each timestep."""

    def __init__(self, asset_name: str, asset_id: str, port_ids: list[str]):
        """Initialize a Pipe object.

        :param str asset_name: The name of the asset.
        :param str asset_id: The unique identifier of the asset.
        :param List[str] port_ids: List of ids of the connected ports.
        """
        super().__init__(asset_name=asset_name, asset_id=asset_id, connected_ports=port_ids)
        # Initialize the default values of the pipe
        self._minor_loss_coefficient = PIPE_DEFAULTS.minor_loss_coefficient
        self._external_temperature = PIPE_DEFAULTS.external_temperature
        self._qheat_external = PIPE_DEFAULTS.qheat_external
        # Define properties of the pipe
        self.length = PIPE_DEFAULTS.length
        self.diameter = PIPE_DEFAULTS.diameter
        self.roughness = PIPE_DEFAULTS.k_value
        self.alpha_value = PIPE_DEFAULTS.alpha_value
        self.solver_asset = SolverPipe(
            name=self.name,
            _id=self.asset_id,
            length=self.length,
            diameter=self.diameter,
            roughness=self.roughness,
        )
        self.output = []

    def _get_diameter(self, esdl_asset: EsdlAssetObject) -> float:
        """Retrieve the diameter of the pipe and convert it if necessary."""
        temp_diameter, property_available = esdl_asset.get_property("innerDiameter", self.diameter)
        if property_available:
            if temp_diameter == 0:
                return self.diameter
            return float(temp_diameter)
        else:
            # Implement DN-conversion
            raise NotImplementedError(
                f"The innderDiamter property is unavailable for {esdl_asset.esdl_asset.name}. \
                    Conversion from DN to diameter is not yet implemented."
            )

    def _get_heat_transfer_coefficient(self, esdl_asset: EsdlAssetObject) -> float:
        """Calculate the heat transfer coefficient of the pipe.

        :param EsdlAssetObject esdl_asset: The ESDL asset object associated with the
                current pipe object.

        :return: The heat transfer coefficient of the pipe [W/(m2 K)]. If the heat transfer
                coefficient cannot be calculated - for example when the material table is
                not specified - , the default alpha value is returned.
        """
        diameters, heat_coefficients = get_thermal_conductivity_table(esdl_asset=esdl_asset)
        if diameters:
            # Create a numpy array of the diameters and heat coefficients
            diameters_np = np.array(diameters)
            heat_coefficients_np = np.array(heat_coefficients)
            # Calculate the heat transfer coefficient from the heat transfer table
            inverse_heat_transfer_coefficient = np.sum(
                calculate_inverse_heat_transfer_coefficient(
                    inner_diameter=diameters_np[:-1],
                    outer_diameter=diameters_np[1:],
                    thermal_conductivity=heat_coefficients_np,
                )
            )
            return 1.0 / float(inverse_heat_transfer_coefficient)
        else:
            return self.alpha_value

    def add_physical_data(self, esdl_asset: EsdlAssetObject) -> None:
        """Method to add physical data to the asset.

        :param EsdlAssetObject esdl_asset: The ESDL asset object associated with the
                current pipe object.
        """
        # Error handling is performed in EsdlAssetObject.get_asset_parameters
        self.length, _ = esdl_asset.get_property(
            esdl_property_name="length", default_value=self.length
        )
        self.roughness, _ = esdl_asset.get_property(
            esdl_property_name="roughness", default_value=self.roughness
        )
        self.roughness = PIPE_DEFAULTS.k_value if self.roughness == 0 else self.roughness
        self.diameter = self._get_diameter(esdl_asset=esdl_asset)

        self.alpha_value = self._get_heat_transfer_coefficient(esdl_asset=esdl_asset)
        prop_dict = {"length": self.length, "diameter": self.diameter, "roughness": self.roughness,
                     "alpha_value": self.alpha_value}
        self.solver_asset.set_physical_properties(physical_properties=prop_dict)

    def set_setpoints(self, setpoints: Dict) -> None:
        """Set the setpoints of the pipe prior to a simulation.

        :param Dict setpoints: The setpoints that should be set for the pipe.
            The keys of the dictionary are the names of the setpoints and the values are the values
        """

    def write_to_output(self) -> None:
        """Write the output of the asset to the output list.

        The output list is a list of dictionaries, where each dictionary
        represents the output of its asset for a specific timestep.
        """
        pass



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Module containing class to store equations."""
import numpy as np


class EquationObject:
    """Class to store equation objects.

    Indices are the location in the matrix of the coefficients
    rhs is the rhs of the matrix.
    """

    indices: np.ndarray
    """Indices of the coefficients in the matrix."""

    coefficients: np.ndarray
    """Coefficients of the equation."""

    rhs: float
    """Right hand side of the equation."""

    def __init__(self) -> None:
        """Constructor of the EquationObject."""
        self.indices = np.array([], dtype=int)
        self.coefficients = np.array([], dtype=float)
        self.rhs = 0.0

    def __len__(self) -> int:
        """Return the number of coefficients in this equation."""
        return len(self.coefficients)

    def to_list(self, length: int) -> list[float]:
        """Method to change the equation object to list which can be stored in the matrix.

        The method creates a list with zero of the length given in the input.
        At the indices of the class the coefficients of the class are stored.
        :param int length: Length of the list to be returned
        :return: a list of the coefficient which can be used in the matrix.
        """
        if length < len(self.indices):
            raise IndexError(
                f"Length of {length} smaller than the available number "
                f"of coefficients {len(self.coefficients)}!"
            )

        coefficient_list: list[float] = [0.0] * length
        for col, coefficient in zip(self.indices, self.coefficients):
            coefficient_list[col] = coefficient
        return coefficient_list



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""NetworkController abstrac."""

from abc import ABC, abstractmethod
import datetime


class NetworkControllerAbstract(ABC):
    """Abstract class for the network controller."""

    @abstractmethod
    def update_setpoints(self, time: datetime.datetime) -> dict:
        """Method to get the controller inputs for the network.

        :param float time: Time step for which to run the controller.
        :return: dict with the key the asset id and the heat demand for that asset.
        """
        pass



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module to read the esdl profiles from an energy system."""
from esdl.profiles.influxdbprofilemanager import InfluxDBProfileManager
from esdl.units.conversion import ENERGY_IN_J, POWER_IN_W, convert_to_unit
import esdl
from esdl.esdl_handler import EnergySystemHandler
from typing import Dict, Tuple
import pandas as pd


def parse_esdl_profiles(esh: EnergySystemHandler) -> Dict[str, pd.DataFrame]:
    """Method to parse the esdl profiles from an energy system.

    This method tries to parse the esdl profiles from an energy system.
    :param esh: EnergySystemHandler with the energy system
    :param start_date: datetime.datetime with the start date
    :param end_date: datetime.datetime with the end date
    :return: dict with the parsed profiles
    """
    data_set = {}
    for profile in esh.get_all_instances_of_type(esdl.InfluxDBProfile):
        data_set[profile.id] = get_data_from_profile(profile)
    return data_set


def get_data_from_profile(esdl_profile: esdl.InfluxDBProfile) -> pd.DataFrame:
    """Method to get the data from the esdl influxdb profile.

    This method tries to get the data from the esdl profile.
    :param esdl_profile: esdl.Profile with the profile
    :return: pandas.DataFrame with the data
    """
    influx_cred_map: Dict[str, Tuple[str, str]] = {}
    profile_host = esdl_profile.host
    ssl_setting = False
    if "https" in profile_host:
        profile_host = profile_host[8:]
        ssl_setting = True
    elif "http" in profile_host:
        profile_host = profile_host[7:]
    # why is this here?
    if esdl_profile.port == 443:
        ssl_setting = True

    influx_host = "{}:{}".format(profile_host, esdl_profile.port)
    if influx_host in influx_cred_map:
        (username, password) = influx_cred_map[influx_host]
    else:
        username = None
        password = None
    time_series_data = InfluxDBProfileManager.create_esdl_influxdb_profile_manager(
        esdl_profile,
        username,
        password,
        ssl_setting,
        ssl_setting,
    )
    # Error check start and end dates of profiles

    # I do not thing this is required since you set it in mapeditor.
    if time_series_data.end_datetime != esdl_profile.endDate:
        raise RuntimeError(
            f"The user input profile end datetime: {esdl_profile.endDate} does not match the end"
            f" datetime in the datbase: {time_series_data.end_datetime} for variable: "
            f"{esdl_profile.field}"
        )
    if time_series_data.start_datetime != esdl_profile.startDate:
        raise RuntimeError(
            f"The user input profile start datetime: {esdl_profile.startDate} does not match the"
            f" start date in the datbase: {time_series_data.start_datetime} for variable: "
            f"{esdl_profile.field}"
        )
    if time_series_data.start_datetime != time_series_data.profile_data_list[0][0]:
        raise RuntimeError(
            f"The profile's variable value for the start datetime: "
            f"{time_series_data.start_datetime} does not match the start datetime of the"
            f" profile data: {time_series_data.profile_data_list[0][0]}"
        )
    if time_series_data.end_datetime != time_series_data.profile_data_list[-1][0]:
        raise RuntimeError(
            f"The profile's variable value for the end datetime: "
            f"{time_series_data.end_datetime} does not match the end datetime of the"
            f" profile data: {time_series_data.profile_data_list[-1][0]}"
        )

    unit = get_unit(esdl_profile)
    dates = [time_stamp for time_stamp, _ in time_series_data.profile_data_list]
    values = [
        convert_to_unit(value, esdl_profile.profileQuantityAndUnit, unit) * esdl_profile.multiplier
        for _, value in time_series_data.profile_data_list
    ]
    data_points = {"date": dates, "values": values}

    return pd.DataFrame(data_points)


def get_unit(profile: esdl.InfluxDBProfile) -> esdl.PhysicalQuantityEnum:
    """Method to get the SI unit of the profile.

    This method tries to get the unit of the profile.
    if the unit is not in the list the same unit as the profile is returned
    :param profile: esdl.InfluxDBProfile with the profile
    :return: esdl.PhysicalQuantityEnum with the unit
    """
    conversion_dict = {
        esdl.PhysicalQuantityEnum.POWER: POWER_IN_W,
        esdl.PhysicalQuantityEnum.ENERGY: ENERGY_IN_J,
    }
    try:
        profile_unit = profile.profileQuantityAndUnit.reference.physicalQuantity
    except AttributeError:
        profile_unit = profile.profileQuantityAndUnit.physicalQuantity
    if profile_unit in conversion_dict:
        return conversion_dict[profile_unit]
    return profile_unit



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module containing abstract BaseNodeItem class."""
from abc import ABC, abstractmethod

from omotes_simulator_core.solver.matrix.equation_object import EquationObject
from omotes_simulator_core.solver.network.assets.base_item import BaseItem
from omotes_simulator_core.solver.matrix.index_core_quantity import index_core_quantity


class BaseNodeItem(ABC):
    """A base class for node items in a network."""

    def __init__(self, name: str, _id: str, number_of_unknowns: int):
        """Initializes the BaseNodeItem object with the given parameters.

        :param int number_of_unknowns: The number of unknown variables for the item.
        :param str name: The name of the node.
        :param str _id: The unique identifier of the node.
        """
        self.name = name
        self.id = _id
        self.number_of_unknowns = number_of_unknowns
        self.matrix_index = 0
        self.prev_sol: list[float] = [0.0] * self.number_of_unknowns

    def set_matrix_index(self, index: int) -> None:
        """Sets the matrix index of the item.

        :param int index: The index of the item in the matrix.
        """
        self.matrix_index = index

    def get_index_matrix(self, property_name: str, use_relative_indexing: bool) -> int:
        """Method to get matrix index of a certain property for a connection point.

        :param str property_name: The property name for which the matrix index is needed.
        :param bool matrix: Whether the index in the matrix should be included.
        :return: The matrix index for the property and connection point.
        """
        return index_core_quantity.get_index_property(
            property_name=property_name, connection_point=0
        ) + (0 if use_relative_indexing else self.matrix_index)

    @abstractmethod
    def get_equations(self) -> list[EquationObject]:
        """Returns the equations for the item.

        :return: The equations for the item.
        :rtype: list[EquationObject]
        """

    @abstractmethod
    def connect_asset(self, asset: BaseItem, connection_point: int) -> None:
        """Connects the asset to the node item.

        :param BaseItem asset: The asset to connect to the node item.
        :param int connection_point: The connection point of the asset to connect to.
        """

    @abstractmethod
    def get_connected_assets(self) -> list[tuple[BaseItem, int]]:
        """Returns the connected assets of the node item.

        :return: The connected assets of the node item.
        :rtype: Tuple[BaseItem, int]
        """



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Define default values and names for assets."""
from dataclasses import dataclass

# Default values
DEFAULT_DIAMETER = 1.2  # [m]
DEFAULT_PRESSURE = 10.0e5  # [bar]
DEFAULT_PRESSURE_DIFFERENCE = 5e5  # [bar]
DEFAULT_TEMPERATURE = 300.0  # [K]
DEFAULT_TEMPERATURE_DIFFERENCE = 30.0  # [K]
DEFAULT_NODE_HEIGHT = 0.0  # [m]
DEFAULT_MASS_FLOW_RATE = 1.0  # [kg/s]
DEFAULT_POWER = 500000.0  # [W]
DEFAULT_MISSING_VALUE = -9999.99  # [-]


@dataclass
class PipeDefaults:
    """Class containing the default values for a pipe.

    :param float k_value: The k value of the pipe [m].
    :param float alpha_value: The alpha value of the pipe [W/(m2 K)].
    :param float minor_loss_coefficient: The minor loss coefficient of the pipe [-].
    :param float external_temperature: The external temperature of the pipe [K].
    :param float qheat_external: The external heat flow of the pipe [W].
    """

    k_value: float = 2e-3
    alpha_value: float = 0.0
    minor_loss_coefficient: float = 0.0
    external_temperature: float = 273.15 + 20.0
    qheat_external: float = 0.0
    length: float = 1.0
    diameter: float = DEFAULT_DIAMETER


@dataclass
class AtesDefaults:
    """Class containing the default values for ATES."""

    aquifer_depth: float = 300.0  # meters
    aquifer_thickness: float = 45.0  # meters
    aquifer_mid_temperature: float = 17.0  # Celcius
    aquifer_net_to_gross: float = 1.0  # percentage
    aquifer_porosity: float = 0.3  # percentage
    aquifer_permeability: float = 10000.0  # mD
    aquifer_anisotropy: float = 4.0  # -
    salinity: float = 10000.0  # ppm
    well_casing_size: float = 13.0  # inch
    well_distance: float = 150.0  # meters
    maximum_flow_charge: float = 200.0  # m3/h
    maximum_flow_discharge: float = 200.0  # m3/h


# Default names
PROPERTY_HEAT_DEMAND = "heat_demand"
PROPERTY_TEMPERATURE_SUPPLY = "temperature_supply"
PROPERTY_TEMPERATURE_RETURN = "temperature_return"
PROPERTY_TEMPERATURE = "temperature"
PROPERTY_PRESSURE_SUPPLY = "pressure_supply"
PROPERTY_PRESSURE_RETURN = "pressure_return"
PROPERTY_PRESSURE = "pressure"
PROPERTY_MASSFLOW = "mass_flow"
PROPERTY_VOLUMEFLOW = "volume_flow"
PROPERTY_THERMAL_POWER = "thermal_power"
PROPERTY_VELOCITY_SUPPLY = "velocity_supply"
PROPERTY_VELOCITY_RETURN = "velocity_return"
PROPERTY_SET_PRESSURE = "set_pressure"
PROPERTY_LENGTH = "length"
PROPERTY_DIAMETER = "diameter"
PROPERTY_ROUGHNESS = "roughness"
PROPERTY_ALPHA_VALUE = "alpha_value"

# Static members
PIPE_DEFAULTS = PipeDefaults()
ATES_DEFAULTS = AtesDefaults()



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Junction classes."""

from typing import Any, List, Optional

from omotes_simulator_core.entities.assets.asset_defaults import (
    DEFAULT_NODE_HEIGHT,
    DEFAULT_PRESSURE,
    DEFAULT_TEMPERATURE,
)
from omotes_simulator_core.solver.network.assets.node import Node


class Junction:
    """Juntion connects two or more assets."""

    def __init__(
        self,
        solver_node: Node,
        pn_bar: float = DEFAULT_PRESSURE,
        tfluid_k: float = DEFAULT_TEMPERATURE,
        height_m: float = DEFAULT_NODE_HEIGHT,
        geodata: Optional[List[Any]] = None,
        name: str = "None",
        in_service: bool = True,
        index: Optional[int] = None,
    ):
        """Initialize a Junction object.

        :param pn_bar: The pressure at the junction [bar], defaults to DEFAULT_PRESSURE
        :type pn_bar: float, optional
        :param tfluid_k: The temperature at the junction [K], defaults to DEFAULT_TEMPERATURE
        :type tfluid_k: float, optional
        :param height_m: The height of the junction [m], defaults to DEFAULT_NODE_HEIGHT
        :type height_m: float, optional

        """
        self.pn_bar = pn_bar
        self.tfluid_k = tfluid_k
        self.height_m = height_m
        self.name = name
        self.index = index
        # Initialize the junction
        self.solver_junction = solver_node

    def write_to_output(self) -> None:
        """Placeholder to write the asset to the output.

        The output list is a list of dictionaries, where each dictionary
        represents the output of its asset for a specific timestep.
        """
        return



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""module containing pipe class."""

from typing import Dict

import numpy as np
from scipy.optimize import root

from omotes_simulator_core.entities.assets.asset_defaults import (
    DEFAULT_MISSING_VALUE,
    PROPERTY_DIAMETER,
    PROPERTY_LENGTH,
    PROPERTY_ROUGHNESS,
    PROPERTY_ALPHA_VALUE
)
from omotes_simulator_core.solver.network.assets.fall_type import FallType
from omotes_simulator_core.solver.utils.fluid_properties import fluid_props


class SolverPipe(FallType):
    """Class to represent a pipe in a network."""

    lambda_loss: float = 0.01
    """ The dimensionless lambda loss value for the pipe. """

    loss_coefficient: float = 0.0
    r""" The loss coefficient of the pipe equal to .. math:: \frac{\lambda L}{2D A^2 g} """

    reynolds_number: float = 0.0
    r""" The Reynolds number of the flow in the pipe .. math:: \frac{v D}{\nu}  """

    alpha_value: float = 0.0
    """ The heat transfer coefficient of the pipe with units [W/m^2/K] """

    ambient_temperature: float = 293.15
    """ The ambient temperature of the pipe with units [K] """

    _use_fluid_capacity: bool = False
    """ A boolean to determine if the fluid capacity is used in the heat transfer calculation. """

    def __init__(
        self,
        name: str,
        _id: str,
        length: float = 1000.0,
        diameter: float = 0.2,
        roughness: float = 0.001,
    ):
        """Constructor of pipe class.

        :param str name: The name of the pipe.
        :param str _id: The unique identifier of the pipe.
        :param float length: The length of the pip [m] with a default value of 1000.0 m.
        :param float diameter: The diameter of the pipe [m] with a default value of 0.2 m.
        :param float roughness: The roughness of the pipe [m] with a default value of 1E-3 m.
        """
        super().__init__(name=name, _id=_id)
        # Set the physical properties of the pipe
        self.length: float = length
        self.diameter: float = diameter
        self.roughness: float = roughness
        # Calculate the area of the pipe
        self.area: float = np.pi * self.diameter**2 / 4

    def set_physical_properties(self, physical_properties: Dict[str, float]) -> None:
        """Method to set the physical properties of the pipe.

        :param physical_properties: dictionary containing the physical properties of the pipe.
        expected properties are: length [m], diameter [m], roughness [m]
        """
        expected_properties = [PROPERTY_LENGTH, PROPERTY_DIAMETER, PROPERTY_ROUGHNESS,
                               PROPERTY_ALPHA_VALUE]

        for expected_property in expected_properties:
            if expected_property not in physical_properties:
                raise ValueError(f"Property {expected_property} is missing in physical_properties")
            if hasattr(self, expected_property):
                setattr(self, expected_property, physical_properties[expected_property])
            else:
                raise ValueError(
                    f"Property {expected_property} is not a valid property " f"of the pipe"
                )
        # Update the area of the pipe
        self.area = np.pi * self.diameter**2 / 4

    def update_loss_coefficient(self) -> None:
        r"""Method to update the loss coefficient of the pipe.

        The loss coefficient is calculated using the following equation:

        .. math:: \lambda \frac{L}{D} \frac{1}{2} \frac{1}{\rho A^2}

        using the mass flow rate to velocity conversion:

        .. math:: v = \frac{\dot{m}}{\rho A}

        with the density of the fluid defined at the first connection point.
        """
        self.calc_lambda_loss()
        density = fluid_props.get_density(
            fluid_props.get_t(
                self.prev_sol[
                    self.get_index_matrix(
                        property_name="internal_energy",
                        connection_point=0,
                        use_relative_indexing=True,
                    )
                ]
            )
        )
        self.loss_coefficient = (
            self.lambda_loss
            * (self.length / self.diameter)
            * (1 / 2)
            * (1 / (self.area**2 * density))
        )

    # TODO: Do we want to implement a dependency on the connection point?
    def calculate_reynolds_number(
        self,
        mass_flow_rate: float = DEFAULT_MISSING_VALUE,
        temperature: float = DEFAULT_MISSING_VALUE,
    ) -> float:
        r"""Method to calculate the Reynolds number of the flow in the pipe.

        The Reynolds number is calculated using the following equation:

        .. math:: \frac{v D}{\nu}

        with the fluid properties defined in the fluid properties module. If
        the mass flow rate and temperature are not provided, the mass flow rate
        and temperature at the first connection point from the previous solution
        are used.

        :param float mass_flow_rate: The mass flow rate of the fluid (kg/s).
        :param float temperature: The temperature of the fluid (K).
        """
        # Retrieve properties from previous solution
        if mass_flow_rate == DEFAULT_MISSING_VALUE:
            mass_flow_rate = self.prev_sol[
                self.get_index_matrix(
                    property_name="mass_flow_rate", connection_point=0, use_relative_indexing=True
                )
            ]
        if temperature == DEFAULT_MISSING_VALUE:
            temperature = fluid_props.get_t(
                self.prev_sol[
                    self.get_index_matrix(
                        property_name="internal_energy",
                        connection_point=0,
                        use_relative_indexing=True,
                    )
                ]
            )
        # Calculate the Reynolds number
        density = fluid_props.get_density(temperature)
        discharge = mass_flow_rate / density
        velocity = discharge / self.area
        return velocity * self.diameter / fluid_props.get_viscosity(temperature)

    def calc_lambda_loss(self) -> None:
        r"""Method to calculate the lambda loss of the pipe.

        For Reynolds numbers lower than 100:
            .. math:: \lambda = 0.64

        For Reynolds numbers between 100 and 2000:
            .. math:: \lambda = \frac{64}{Re}

        For Reynolds numbers higher than 2000 use the Tkachenko and Mileikovsky (2020) correlation:
            .. math::

                f = \left( \frac{8.128943 + A_1}{8.128943 A_0 - 0.86859209 A_1 \ln \left(
                        \frac{A_1}{3.7099535 Re} \right)} \right)^2

        with the following coefficients:

            .. math::

                    A_0 = -0.79638 \ln \left( \frac{\frac{\epsilon}{D}}{8.208}
                        + \frac{7.3357}{Re} \right)

                    A_1 = Re \left( \frac{\epsilon}{D} \right) + 9.3120665 A_0

        """
        # Update the Reynolds number
        self.reynolds_number = self.calculate_reynolds_number()
        # Determine the loss
        if self.reynolds_number < 100:
            self.lambda_loss = 0.64
        elif self.reynolds_number < 2000:
            self.lambda_loss = 64 / self.reynolds_number
        elif self.reynolds_number < 4000:
            # Interpolate between f calculated with Reynolds = 2000.0 and f calculated with
            # Reynolds = 4000
            f_lower = 64.0 / 2000.0
            f_upper = self._calculate_explicit_friction_factor(
                reynolds_number=4000.0
            )  # self._colebrook_white(reynolds_number=4000.0)
            self.lambda_loss = f_lower + ((f_upper - f_lower) / 2000.0) * (
                self.reynolds_number - 2000.0
            )
        else:
            self.lambda_loss = self._calculate_explicit_friction_factor()  # self._colebrook_white()

    def _calculate_explicit_friction_factor(self, reynolds_number: float | None = None) -> float:
        r"""Method to calculate the explicit friction factor of the pipe.

        The explicit friction factor is calculated with the Tkaenko and Mileikovsky (2020)
        correlation:

        .. math::

            f = \left( \frac{8.128943 + A_1}{8.128943 A_0 - 0.86859209 A_1 \ln \left(
                    \frac{A_1}{3.7099535 Re} \right)} \right)^2

        with the following coefficients:

        .. math::

                A_0 = -0.79638 \ln \left( \frac{\frac{\epsilon}{D}}{8.208}
                    + \frac{7.3357}{Re} \right)

                A_1 = Re \left( \frac{\epsilon}{D} \right) + 9.3120665 A_0


        :param float reynolds_number: The Reynolds number of the flow in the pipe.
        :return: float, the friction factor.
        """
        # Check existence of Reynolds number
        if reynolds_number is None:
            reynolds_number = self.reynolds_number
        # Calculate the coefficients
        A0 = -0.79638 * np.log((self.roughness / self.diameter) / 8.208 + 7.3357 / reynolds_number)
        A1 = reynolds_number * (self.roughness / self.diameter) + 9.3120665 * A0

        # Calculate the friction factor
        return float(
            (
                (8.128943 + A1)
                / (8.128943 * A0 - 0.86859209 * A1 * np.log(A1 / (3.7099535 * reynolds_number)))
            )
            ** 2
        )

    def _colebrook_white_objective(self, lambda_guess: float) -> float:
        r"""Root function for the Colebrook-White equation.

        Solves the following optimization objective:

        .. math::

            - 2 \log \left( \frac{\epsilon}{D 3.71} +   \frac{2.51}{Re \sqrt{\lambda}}
            \right) - \frac{1}{\sqrt{\lambda}} = 0

        :param float lambda_val: The friction factor.
        :return: float, the friction factor.
        """
        # Calculate the friction factor using the Colebrook-White equation
        return float(
            (
                -2
                * np.log10(
                    (2.51 / (self.reynolds_number * np.sqrt(lambda_guess)))
                    + (self.roughness / (3.7 * self.diameter))
                )
                - 1.0 / np.sqrt(lambda_guess)
            ).item()
        )

    def _colebrook_white(
        self,
        lambda_guess: float = 0.001,
        convergence_limit: float = 1e-4,
        reynolds_number: float = DEFAULT_MISSING_VALUE,
    ) -> float:
        r"""Method to iteratively calculate the Colebrook-White friction factor.

        :param float lambda_guess: The initial guess for the 1/\sqrt{friction factor}.
        :param float convergence_limit: The convergence limit for the iterative calculation.
        :param float reynolds_number: The Reynolds number of the flow in the pipe.
        :return: float, the friction factor.
        """
        if reynolds_number == DEFAULT_MISSING_VALUE:
            reynolds_number = self.reynolds_number
        # Iterative calculation
        return float(
            root(
                self._colebrook_white_objective, lambda_guess, method="hybr", tol=convergence_limit
            ).x[0]
        )

    def _calculate_graetz_number(
        self, temperature: float, mass_flow_rate: float = DEFAULT_MISSING_VALUE
    ) -> float:
        r"""Calculate the Graetz number of the flow in the pipe.

        If the mass flow rate is not provided, the mass flow rate from the previous solution is
         used.

        The Graetz number is calculated using the following equation:

        .. math:: \text{Gz} = \frac{\alpha A}{D v}

        If the velocity is zero, a large number is returned to indicate a low heat
        transfer coefficient.

        :param float temperature: The temperature of the fluid.
        :param float mass_flow_rate: The mass flow rate of the fluid.
        :return: float, the Graetz number.
        """
        # Calculate the Reynolds number
        reynolds_number = self.calculate_reynolds_number(
            mass_flow_rate=mass_flow_rate, temperature=temperature
        )

        # Calculate the Prandtl number
        prandtl_number = self.calculate_prandtl_number(temperature=temperature)

        # Check velocity
        if abs(reynolds_number) >= 1e-6:
            return self.diameter / self.length * reynolds_number * prandtl_number
        else:
            # Return a large number so that the heat transfer coefficient is very low
            return 10.0

    def calculate_prandtl_number(self, temperature: float = DEFAULT_MISSING_VALUE) -> float:
        r"""Calculate the Prandtl number of the fluid.

        The Prandtl number is calculated using the following equation:

        .. math:: \text{Pr} = \frac{\mu}{\alpha}

        with the dynamic viscosity and thermal diffusivity of the fluid. The thermal diffusivity is
        calculated as:

        .. math:: \alpha = \frac{k}{\rho c_p}

        which is the ratio of the dynamic viscosity to the thermal diffusivity of the fluid.

        :param float temperature: The temperature of the fluid.
        :return: float, the Prandtl number.
        """
        # Check the temperature
        if temperature == DEFAULT_MISSING_VALUE:
            temperature = fluid_props.get_t(
                self.prev_sol[
                    self.get_index_matrix(
                        property_name="internal_energy",
                        connection_point=0,
                        use_relative_indexing=True,
                    )
                ]
            )

        # Calculate the thermal diffusivity
        thermal_diffusivity = fluid_props.get_thermal_conductivity(temperature) / (
            fluid_props.get_density(temperature) * fluid_props.get_heat_capacity(temperature)
        )
        # Calculate the Prandtl number
        return fluid_props.get_viscosity(temperature) / thermal_diffusivity

    def _calculate_heat_transfer_coefficient_fluid(
        self, temperature: float, mass_flow_rate: float
    ) -> float:
        r"""Calculate the heat transfer coefficient of the fluid.

        The heat transfer coefficient is calculated using the Nusselt number and the thermal
        conductivity of the fluid.

        If the Reynolds number is less than 1e4, the Nusselt number is calculated using the
        Graetz number. For a Graetz number greater than 0.1, the Nusselt number is set to 3.66.
        For a Graetz number less than 0.1, the Nusselt number is calculated using the following
        equation:

        .. math:: Nu = 1.62 \cdot \text{Gz}^{-1/3}

        Otherwise, the Nusselt number is calculated using the Reynolds number and

        .. math:: Nu = 0.023 \cdot \text{Re}^{0.8} \cdot \text{Pr}^{0.33}

        where Pr is the Prandtl number of the fluid as:

        .. math:: \text{Pr} = \frac{\mu}{\alpha}

        The heat transfer coefficient is then calculated using the following equation:

        .. math:: \alpha = Nu \cdot k / D

        Source: R. Mudde, 'Fysische transportverschijnselen', 1998 p. 133 - p.134

        :param float temperature: The temperature of the fluid.
        :param float mass_flow_rate: The mass flow rate of the fluid.
        """
        # Calculate the Reynolds number
        reynolds_number = self.calculate_reynolds_number(
            mass_flow_rate=mass_flow_rate, temperature=temperature
        )
        # Determine the Nusselt number
        if reynolds_number < 1e4:
            # Laminar flow assumptions
            graetz_number = self._calculate_graetz_number(
                temperature=temperature, mass_flow_rate=mass_flow_rate
            )
            # TODO: Discuss Graetz is defined for (1/Gz) < 0.05 and (1/Gz) > 0.1
            if (1 / graetz_number) >= 0.1:
                nusselt_number = 3.66
            else:
                nusselt_number = 1.62 * graetz_number ** (-1 / 3)
        else:
            prandtl_number = self.calculate_prandtl_number(temperature=temperature)
            nusselt_number = 0.023 * reynolds_number**0.8 * prandtl_number**0.33
        # Calculate the heat transfer coefficient
        return nusselt_number * fluid_props.get_thermal_conductivity(temperature) / self.diameter

    def _determine_inflow_temperature(self) -> tuple[float, float]:
        """Determine the inflow temperature and mass flow rate of the pipe.

        The inflow temperature is determined by looking at the flow direction in the pipe
        """
        mass_flow_rate = self.prev_sol[
            self.get_index_matrix(
                property_name="mass_flow_rate", connection_point=0, use_relative_indexing=True
            )
        ]
        # Determine the flow direction
        if mass_flow_rate < 0:
            # Flow from connection point 1 to connection point 0
            tin = fluid_props.get_t(
                self.prev_sol[
                    self.get_index_matrix(
                        property_name="internal_energy",
                        connection_point=1,
                        use_relative_indexing=True,
                    )
                ]
            )
            mass_flow_rate = abs(mass_flow_rate)
        elif mass_flow_rate > 0:
            # Flow from connection point 0 to connection point 1
            tin = fluid_props.get_t(
                self.prev_sol[
                    self.get_index_matrix(
                        property_name="internal_energy",
                        connection_point=0,
                        use_relative_indexing=True,
                    )
                ]
            )
        else:
            # No flow
            tin = self.ambient_temperature
        return tin, float(mass_flow_rate)

    def _calculate_total_heat_transfer_coefficient(
        self, temperature: float, mass_flow_rate: float
    ) -> float:
        r"""Calculate the total heat transfer coefficient of the pipe.

        The total heat transfer coefficient is calculated using the following equation:

        .. math::

            \frac{1}{\alpha} = \frac{1}{\alpha_{\text{fluid}}} +
            \frac{1}{\alpha_{\text{wall}}}

        where the heat transfer coefficient of the fluid is calculated using the
        '_calculate_heat_transfer_coefficient_fluid' method and the heat transfer coefficient of the
        wall is the alpha value of the pipe.

        :param float temperature: The temperature of the fluid.
        :param float mass_flow_rate: The mass flow rate of the fluid.
        :return: float, the total heat transfer coefficient.
        """
        if self._use_fluid_capacity:
            # Determine the heat transfer coefficient including fluid properties
            heat_transfer_coefficient_fluid = self._calculate_heat_transfer_coefficient_fluid(
                temperature=temperature,
                mass_flow_rate=mass_flow_rate,
            )
            if self.alpha_value > 0.0:
                return 1 / (1 / heat_transfer_coefficient_fluid + 1 / self.alpha_value)
            else:
                return heat_transfer_coefficient_fluid
        else:
            return self.alpha_value

    def _calculate_total_heat_loss(self, tin: float, mass_flow_rate: float) -> float:
        r"""Calculate the total heat loss of the pipe.

        The heat loss is calculated with the following formula:
        .. math::

        \dot{m} c_p \left(T_{in} - T_{out}\right) \left(1 -
        \exp{\frac{-\alpha \pi D L}{\dot{m} c_p}})\right

        :param float tin: The temperature of the fluid at the inlet of the pipe.
        :param float mass_flow_rate: The mass flow rate of the fluid in the pipe.
        :return: float, the total heat loss of the pipe.
        """
        if mass_flow_rate == 0:
            return 0.0
        cp = fluid_props.get_heat_capacity(tin)
        heat_loss = (
            mass_flow_rate
            * cp
            * (tin - self.ambient_temperature)
            * (
                1
                - np.exp(
                    -self.alpha_value * np.pi * self.diameter * self.length / (mass_flow_rate * cp)
                )
            )
        )
        return float(heat_loss)

    def update_heat_supplied(self) -> None:
        """Calculate the heat supplied by the pipe.

        Definition: positive if heat is supplied to the fluid, negative if heat is extracted from
        the fluid.

        """
        tin, mass_flow_rate = self._determine_inflow_temperature()

        self._calculate_total_heat_transfer_coefficient(
            temperature=tin, mass_flow_rate=mass_flow_rate
        )
        self.heat_supplied = -self._calculate_total_heat_loss(
            tin=tin, mass_flow_rate=mass_flow_rate
        )



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module containing the node class."""
import numpy as np

from omotes_simulator_core.solver.matrix.equation_object import EquationObject
from omotes_simulator_core.solver.network.assets.base_item import BaseItem
from omotes_simulator_core.solver.network.assets.base_node_item import BaseNodeItem
from omotes_simulator_core.solver.utils.fluid_properties import fluid_props


class Node(BaseNodeItem):
    """
    A class to represent a node in a network.

    This class inherits from the BaseItem class and implements the methods to generate the
    equations for the node.

    Attributes
    ----------
    connected_assets : list[list[BaseItem, int]]
        A list of lists that store the asset objects and the connection point indices that
        are connected to the node.

    Methods
    -------
    connect_asset(asset: BaseItem, connection_point: int)
        Connects an asset object to the node at the given connection point index.
    get_equations() -> list[EquationObject]
        Returns a list of EquationObjects that represent the equations for the node.
    get_energy_equations() -> EquationObject
        Returns an EquationObject that represents the energy balance equation for the node.
    get_node_cont_equation() -> EquationObject
        Returns an EquationObject that represents the mass flow rate continuity equation for
        the node.
    get_discharge_equation() -> EquationObject
        Returns an EquationObject that represents the discharge equation for the node.
    get_pressure_set_equation() -> EquationObject
        Returns an EquationObject that represents the pressure set equation for the node.
    set_temperature_equation() -> EquationObject
        Returns an EquationObject that represents the temperature set equation for the node.
    get_energy_equation() -> EquationObject
        Returns an EquationObject that represents the energy equation for the node.
    is_connected() -> bool
        Returns True if the node is connected to any asset, False otherwise.
    """

    def __init__(
        self,
        name: str,
        _id: str,
        number_of_unknowns: int = 3,
        height: float = 0.0,
        initial_temperature: float = 273.15,
        set_pressure: float = 10000.0,
    ):
        """Initializes the Node object with the given parameters.

        :param uuid.UUID name: The unique identifier of the node.
        :param int, optional number_of_unknowns: The number of unknown variables for the node.
        The default is 3.
        """
        super().__init__(name=name, _id=_id, number_of_unknowns=number_of_unknowns)
        self.connected_assets: list[tuple[BaseItem, int]] = []
        self.height = height
        self.initial_temperature = initial_temperature
        self.set_pressure = set_pressure

    def connect_asset(self, asset: BaseItem, connection_point: int) -> None:
        """Connects an asset object at the given connection point to the node .

        :param BaseAsset asset: The asset object that is connected to the node.
        :param int connection_point: The connection point for the asset which is connected to this
            node.
        :return:
        """
        if connection_point > (asset.number_of_connection_point - 1):
            raise ValueError(
                f"Connection point {connection_point} does not exist on asset {asset.name}."
            )
        # Check if the asset is already connected to the node at the given connection point
        if set((asset, connection_point)) in set(self.connected_assets):
            raise ValueError(
                f"Asset {asset.name} is already connected to node {self.name} at connection"
                + f" point {connection_point}."
            )
        else:
            # Connect the asset to the node
            self.connected_assets.append((asset, connection_point))

    def get_equations(self) -> list[EquationObject]:
        """Returns a list of EquationObjects that represent the equations for the node.

        The equations are:

        - Mass flow rate continuity equation
        - Energy balance equation
        - Discharge equation
        :return: list[EquationObject]
            A list of EquationObjects that contain the indices, coefficients, and right-hand side
            values of the equations.
        """
        # Check connection
        if not self.is_connected():
            raise ValueError(f"Node {self.name} is not connected to any asset.")
        # Construct equation object
        equations = [
            self.get_node_cont_equation(),
            self.get_energy_equations(),
            self.get_discharge_equation(),
        ]
        return equations

    def get_energy_equations(self) -> EquationObject:
        """Returns an EquationObject that represents the energy balance equation for the node.

        When the mass flow rate of all connected components is smaller or equal 0.
        Then the node will pre-scribe its temperature otherwise it will give
        an equation where the sum of mass flow rat times specific internal energy is zero.


        :return: EquationObject An EquationObject that contains the indices, coefficients,
            and right-hand side value of the equation.
        """
        flows = np.array(
            [
                asset.prev_sol[
                    asset.get_index_matrix(
                        "mass_flow_rate", asset_connection_point, use_relative_indexing=True
                    )
                ]
                for asset, asset_connection_point in self.connected_assets
            ]
        )

        if all(flows >= 0.0) or all(flows <= 0.0):
            return self.set_temperature_equation()
        else:
            return self.get_energy_equation()

    def get_node_cont_equation(self) -> EquationObject:
        """Returns an EquationObject that represents the mass continuity equation for the node.

        :return: EquationObject
            An EquationObject that contains the indices, coefficients, and right-hand side value
            of the equation.
        """
        equation_object = EquationObject()
        equation_object.indices = np.array(
            [self.get_index_matrix(property_name="mass_flow_rate", use_relative_indexing=False)]
        )
        equation_object.coefficients = np.array([1.0])
        equation_object.rhs = 0.0
        for asset, asset_connection_point in self.connected_assets:
            equation_object.indices = np.append(
                equation_object.indices,
                [
                    asset.get_index_matrix(
                        property_name="mass_flow_rate",
                        connection_point=asset_connection_point,
                        use_relative_indexing=False,
                    )
                ],
            )
            equation_object.coefficients = np.append(equation_object.coefficients, [1.0])
        return equation_object

    def get_discharge_equation(self) -> EquationObject:
        """Returns an EquationObject that represents the discharge is zero equation for the node.

        :return: EquationObject
            An EquationObject that contains the indices, coefficients, and right-hand side
            value of the equation.
        """
        equation_object = EquationObject()
        equation_object.indices = np.array(
            [self.get_index_matrix(property_name="mass_flow_rate", use_relative_indexing=False)]
        )
        equation_object.coefficients = np.array([1.0])
        equation_object.rhs = 0.0
        return equation_object

    def get_pressure_set_equation(self) -> EquationObject:
        """Returns an EquationObject that sets the pressure of the node to a pre-defined value.

        :return: EquationObject
            An EquationObject that contains the indices, coefficients, and right-hand side
            value of the equation.
        """
        equation_object = EquationObject()
        equation_object.indices = np.array(
            [self.get_index_matrix(property_name="pressure", use_relative_indexing=False)]
        )
        equation_object.coefficients = np.array([1.0])
        equation_object.rhs = self.set_pressure
        return equation_object

    def set_temperature_equation(self) -> EquationObject:
        """Returns an EquationObject that sets the temperature of the node to a pre-defined value.

        :return: EquationObject
            An EquationObject that contains the indices, coefficients, and right-hand side
            value of the equation.
        """
        equation_object = EquationObject()
        equation_object.indices = np.array(
            [self.get_index_matrix(property_name="internal_energy", use_relative_indexing=False)]
        )
        equation_object.coefficients = np.array([1.0])
        equation_object.rhs = fluid_props.get_ie(self.initial_temperature)
        return equation_object

    def get_energy_equation(self) -> EquationObject:
        """Returns an EquationObject that represents the energy equation for the node.

        :return: EquationObject
            An EquationObject that contains the indices, coefficients, and right-hand side
            value of the equation
        """
        equation_object = EquationObject()
        equation_object.indices = np.array(
            [
                self.get_index_matrix(property_name="mass_flow_rate", use_relative_indexing=False),
                self.get_index_matrix(property_name="internal_energy", use_relative_indexing=False),
            ]
        )
        # Be aware that the coefficients are in reverse order
        equation_object.coefficients = np.array(self.prev_sol)[
            (equation_object.indices - self.matrix_index)[::-1]
        ]
        equation_object.rhs = float(np.prod(equation_object.coefficients))
        # Extend the equation_object with the indices and coefficients of the connected assets
        for asset, asset_connection_id in self.connected_assets:
            # Extended asset indices
            extra_indices = np.array(
                [
                    asset.get_index_matrix(
                        "mass_flow_rate", asset_connection_id, use_relative_indexing=False
                    ),
                    asset.get_index_matrix(
                        "internal_energy", asset_connection_id, use_relative_indexing=False
                    ),
                ]
            )
            # Extend the indices and coefficients of the equation object
            equation_object.indices = np.append(
                equation_object.indices,
                extra_indices,
            )
            # Extend the coefficients of the equation object
            prev_sol = np.array(asset.prev_sol)
            equation_object.coefficients = np.append(
                equation_object.coefficients,
                prev_sol[(extra_indices - asset.matrix_index)[::-1]],
            )
            # Extend the right-hand side of the equation object
            equation_object.rhs += np.prod(prev_sol[extra_indices - asset.matrix_index])
        return equation_object

    def is_connected(self) -> bool:
        """Returns True if the node is connected to any asset, False otherwise.

        :return: bool
            A boolean value that indicates whether the node is connected or not.
        """
        return len(self.connected_assets) > 0

    def get_connected_assets(self) -> list[tuple[BaseItem, int]]:
        """Returns the connected assets of the node.

        :return: Tuple[BaseItem, int]
            A tuple of the connected asset and the connection point index.
        """
        return self.connected_assets



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""demandCluster class."""
from typing import Dict

from omotes_simulator_core.entities.assets.asset_abstract import AssetAbstract
from omotes_simulator_core.entities.assets.asset_defaults import (
    DEFAULT_DIAMETER,
    DEFAULT_POWER,
    DEFAULT_PRESSURE,
    DEFAULT_TEMPERATURE,
    DEFAULT_TEMPERATURE_DIFFERENCE,
    PROPERTY_HEAT_DEMAND,
    PROPERTY_TEMPERATURE_RETURN,
    PROPERTY_TEMPERATURE_SUPPLY,
)
from omotes_simulator_core.entities.assets.esdl_asset_object import EsdlAssetObject
from omotes_simulator_core.entities.assets.utils import (
    heat_demand_and_temperature_to_mass_flow,
)
from omotes_simulator_core.solver.network.assets.production_asset import ProductionAsset


class DemandCluster(AssetAbstract):
    """A DemandCluster represents an asset that consumes heat."""

    def __init__(self, asset_name: str, asset_id: str, port_ids: list[str]):
        """Initialize a DemandCluster object.

        :param str asset_name: The name of the asset.
        :param str asset_id: The unique identifier of the asset.
        :param List[str] port_ids: List of ids of the connected ports.
        """
        super().__init__(asset_name=asset_name, asset_id=asset_id, connected_ports=port_ids)

        self._internal_diameter = DEFAULT_DIAMETER
        self.temperature_supply = DEFAULT_TEMPERATURE
        self.temperature_return: float = DEFAULT_TEMPERATURE - DEFAULT_TEMPERATURE_DIFFERENCE
        self.temperature_return_target = self.temperature_return
        self.pressure_input = DEFAULT_PRESSURE
        self.thermal_power_allocation = DEFAULT_POWER
        self.mass_flowrate = 0
        self.solver_asset = ProductionAsset(name=self.name, _id=self.asset_id)
        # Output list
        self.output: list = []

    def set_setpoints(self, setpoints: Dict) -> None:
        """Placeholder to set the setpoints of an asset prior to a simulation.

        :param Dict setpoints: The setpoints that should be set for the asset.
            The keys of the dictionary are the names of the setpoints and the values are the values
        """
        # Default keys required
        necessary_setpoints = {
            PROPERTY_TEMPERATURE_SUPPLY,
            PROPERTY_TEMPERATURE_RETURN,
            PROPERTY_HEAT_DEMAND,
        }
        # Dict to set
        setpoints_set = set(setpoints.keys())
        # Check if all setpoints are in the setpoints
        if not necessary_setpoints.issubset(setpoints_set):
            # Print missing setpoints
            raise ValueError(
                f"The setpoints {necessary_setpoints.difference(setpoints_set)} are missing."
            )
        self.thermal_power_allocation = -setpoints[PROPERTY_HEAT_DEMAND]
        self.temperature_return_target = setpoints[PROPERTY_TEMPERATURE_RETURN]
        self.temperature_supply = setpoints[PROPERTY_TEMPERATURE_SUPPLY]
        adjusted_mass_flowrate = heat_demand_and_temperature_to_mass_flow(
            self.thermal_power_allocation, self.temperature_supply, self.temperature_return_target
        )
        self.solver_asset.supply_temperature = self.temperature_supply
        self.solver_asset.mass_flow_rate_set_point = adjusted_mass_flowrate  # type: ignore

    def add_physical_data(self, esdl_asset: EsdlAssetObject) -> None:
        """Method to add physical data to the asset.

        :param EsdlAssetObject esdl_asset: The esdl asset object to add the physical data from.
         :return:
        """

    def write_to_output(self) -> None:
        """Placeholder to write the asset to the output.

        The output list is a list of dictionaries, where each dictionary
        represents the output of its asset for a specific timestep.
        """
        pass



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""ProductionCluster class."""
from typing import Dict

from omotes_simulator_core.entities.assets.asset_abstract import AssetAbstract
from omotes_simulator_core.entities.assets.asset_defaults import (
    DEFAULT_NODE_HEIGHT,
    DEFAULT_PRESSURE,
    DEFAULT_TEMPERATURE,
    DEFAULT_TEMPERATURE_DIFFERENCE,
    PROPERTY_HEAT_DEMAND,
    PROPERTY_SET_PRESSURE,
    PROPERTY_TEMPERATURE_RETURN,
    PROPERTY_TEMPERATURE_SUPPLY,
)
from omotes_simulator_core.entities.assets.esdl_asset_object import EsdlAssetObject
from omotes_simulator_core.entities.assets.utils import (
    heat_demand_and_temperature_to_mass_flow,
)
from omotes_simulator_core.solver.network.assets.production_asset import ProductionAsset


class ProductionCluster(AssetAbstract):
    """A ProductionCluster represents an asset that produces heat."""

    thermal_production_required: float | None
    """The thermal production required by the asset [W]."""

    temperature_supply: float
    """The supply temperature of the asset [K]."""

    temperature_return: float
    """The return temperature of the asset [K]."""

    pressure_supply: float
    """The supply pressure of the asset [Pa]."""

    control_mass_flow: bool
    """Flag to indicate whether the mass flow rate is controlled.
    If True, the mass flow rate is controlled. If False, the mass flow rate is not controlled
    and the pressure is predescribed.
    """

    controlled_mass_flow: float | None
    """The controlled mass flow of the asset [kg/s]."""

    def __init__(self, asset_name: str, asset_id: str, port_ids: list[str]):
        """Initialize a ProductionCluster object.

        :param str asset_name: The name of the asset.
        :param str asset_id: The unique identifier of the asset.
        :param List[str] port_ids: List of ids of the connected ports.
        """
        super().__init__(asset_name=asset_name, asset_id=asset_id, connected_ports=port_ids)
        self.height_m = DEFAULT_NODE_HEIGHT
        # DemandCluster thermal and mass flow specifications
        self.thermal_production_required = None
        self.temperature_supply = DEFAULT_TEMPERATURE
        self.temperature_return = DEFAULT_TEMPERATURE - DEFAULT_TEMPERATURE_DIFFERENCE
        # DemandCluster pressure specifications
        self.pressure_supply = DEFAULT_PRESSURE
        self.control_mass_flow = False
        # Controlled mass flow
        self.controlled_mass_flow = None
        self.solver_asset = ProductionAsset(
            name=self.name,
            _id=self.asset_id,
            pre_scribe_mass_flow=False,
            set_pressure=self.pressure_supply,
        )

    def add_physical_data(self, esdl_asset: EsdlAssetObject) -> None:
        """Method to add physical data to the asset.

        :param EsdlAssetObject esdl_asset: The ESDL asset object containing the physical data.
        :return:
        """

    def _set_supply_temperature(self, temperature_supply: float) -> None:
        """Set the supply temperature of the asset.

        :param float temperature_supply: The supply temperature of the asset.
            The temperature should be supplied in Kelvin.
        """
        # Set the temperature of the circulation pump mass flow
        self.temperature_supply = temperature_supply
        self.solver_asset.supply_temperature = self.temperature_supply

    def _set_return_temperature(self, temperature_return: float) -> None:
        """Set the return temperature of the asset.

        :param float temperature_return: The return temperature of the asset.
            The temperature should be supplied in Kelvin.
        """
        # Set the return temperature of the asset
        self.temperature_return = temperature_return

    def _set_heat_demand(self, heat_demand: float) -> None:
        """Set the heat demand of the asset.

        :param float heat_demand: The heat demand of the asset.
            The heat demand should be supplied in Watts.
        """
        # Calculate the mass flow rate
        self.controlled_mass_flow = heat_demand_and_temperature_to_mass_flow(
            thermal_demand=heat_demand,
            temperature_supply=self.temperature_supply,
            temperature_return=self.temperature_return,
        )

        # Check if the mass flow rate is positive
        if self.controlled_mass_flow < 0.0:
            raise ValueError(
                f"The mass flow rate {self.controlled_mass_flow} of the asset {self.name}"
                + " is negative."
            )
        else:
            # Set the mass flow rate of the control valve
            self.solver_asset.mass_flow_rate_set_point = self.controlled_mass_flow  # type: ignore

    def _set_pressure_or_mass_flow_control(self, pressure_supply: bool) -> None:
        """Set the asset to predescribe either the pressure or the mass flow rate.

        :param bool pressure_supply: True when the pressure needs to be set
        """
        if pressure_supply:
            self.solver_asset.pre_scribe_mass_flow = False  # type: ignore
        else:
            self.solver_asset.pre_scribe_mass_flow = True  # type: ignore

    def set_pressure_supply(self, pressure_supply: float) -> None:
        """Set the supply pressure of the asset.

        :param float pressure_supply: The supply pressure of the asset.
            The pressure should be supplied in Pascal.
        """
        # Check if the pressure is positive
        if pressure_supply < 0.0:
            raise ValueError(
                f"The pressure {pressure_supply} of the asset {self.name} can not be negative."
            )
        # Set the supply pressure of the asset
        self.pressure_supply = pressure_supply
        # Set the pressure of the solver asset
        self.solver_asset.set_pressure = self.pressure_supply  # type: ignore

    def set_setpoints(self, setpoints: Dict) -> None:
        """Set the setpoints of the asset.

        :param Dict setpoints: The setpoints of the asset in a dictionary,
            as "property_name": value pairs.

        """
        # Default keys required
        necessary_setpoints = {
            PROPERTY_TEMPERATURE_SUPPLY,
            PROPERTY_TEMPERATURE_RETURN,
            PROPERTY_HEAT_DEMAND,
            PROPERTY_SET_PRESSURE,
        }
        # Dict to set
        setpoints_set = set(setpoints.keys())
        # Check if all setpoints are in the setpoints
        if necessary_setpoints.issubset(setpoints_set):
            # Set the setpoints
            self._set_pressure_or_mass_flow_control(setpoints[PROPERTY_SET_PRESSURE])
            self._set_supply_temperature(setpoints[PROPERTY_TEMPERATURE_SUPPLY])
            self._set_return_temperature(setpoints[PROPERTY_TEMPERATURE_RETURN])
            self._set_heat_demand(setpoints[PROPERTY_HEAT_DEMAND])
        else:
            # Print missing setpoints
            raise ValueError(
                f"The setpoints {necessary_setpoints.difference(setpoints_set)} are missing."
            )

    def update(self) -> None:
        """Update the asset properties to the results from the previous (timestep) simulation.

        Sets the values of the supply temperature, return temperature and heat demand
        to the values of the previous simulation. In addition, the mass flow rate is set
        to the value of the previous simulation.
        """

    def write_to_output(self) -> None:
        """Placeholder to write the asset to the output.

        The output list is a list of dictionaries, where each dictionary
        represents the output of its asset for a specific timestep.
        """
        pass



#  Copyright (c) 2024. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Module containing the classes for the controller."""


class AssetControllerAbstract:
    """Abstract class for the asset controller."""

    def __init__(self, name: str, identifier: str):
        """Constructor for the asset controller.

        :param str name: Name of the controller object.
        :param str identifier: Unique identifier of the controller asset.
        """
        self.name = name
        self.id = identifier



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module containing classes to be able to interact with esdl objects."""
import logging
from typing import Any, Tuple, Type

import pandas as pd
from esdl import esdl
from omotes_simulator_core.entities.utility.influxdb_reader import get_data_from_profile
from omotes_simulator_core.adapter.transforms.transform_utils import PortType, sort_ports, Port

logger = logging.getLogger(__name__)


class EsdlAssetObject:
    """
    Class to hold an esdl asset and convert it to local class objects.

    Conversion is done based on the classes in the CONVERSION_DICT.
    """

    esdl_asset: esdl.Asset

    def __init__(self, asset: esdl.Asset) -> None:
        """
        Constructor for EsdlAssetObject class.

        :param asset, esdl.Asset: PyEsdl Asset object
        """
        self.esdl_asset = asset

    def get_id(self) -> str:
        """Get the id of the asset."""
        return str(self.esdl_asset.id)

    def get_property(self, esdl_property_name: str, default_value: Any) -> Tuple[Any, bool]:
        """Get property value from the esdl_asset based on the "ESDL" name.

        :return: Tuple with the value of the property and a boolean indicating whether the property
        was found in the esdl_asset.
        """
        try:
            return getattr(self.esdl_asset, esdl_property_name), True
        except AttributeError:
            return default_value, False

    def get_profile(self) -> pd.DataFrame:
        """Get the profile of the asset."""
        for esdl_port in self.esdl_asset.port:
            if esdl_port.profile:
                return get_data_from_profile(esdl_port.profile[0])
        raise ValueError(f"No profile found for asset: {self.esdl_asset.name}")

    def get_supply_temperature(self, port_type: str) -> float:
        """Get the temperature of the port."""
        for esdl_port in self.esdl_asset.port:
            if isinstance(esdl_port, self.get_port_type(port_type)):
                return get_supply_temperature(esdl_port)
        raise ValueError(f"No port found with type: {port_type} for asset: {self.esdl_asset.name}")

    def get_return_temperature(self, port_type: str) -> float:
        """Get the temperature of the port."""
        for esdl_port in self.esdl_asset.port:
            if isinstance(esdl_port, self.get_port_type(port_type)):
                return get_return_temperature(esdl_port)
        raise ValueError(f"No port found with type: {port_type} for asset: {self.esdl_asset.name}")

    def get_port_ids(self) -> list[str]:
        """Returns a sorted list of the port ids of the asset."""
        list_of_ports = sort_ports(
            [
                Port(
                    port.id,
                    port.name,
                    PortType.IN if isinstance(port, esdl.InPort) else PortType.OUT,
                )
                for port in self.esdl_asset.port
            ]
        )
        return list_of_ports

    def get_port_type(self, port_type: str) -> Type[esdl.Port]:
        """Get the port type of the port."""
        if port_type == "In":
            return esdl.InPort  # type: ignore [no-any-return]
        elif port_type == "Out":
            return esdl.OutPort  # type: ignore [no-any-return]
        else:
            raise ValueError(f"Port type not recognized: {port_type}")

    def get_marginal_costs(self) -> float:
        """Get the marginal costs of the asset."""
        if self.esdl_asset.costInformation is None:
            logger.warning(
                f"No cost information found for asset, Marginal costs set to 0 for: "
                f"{self.esdl_asset.name}"
            )
            return 0
        if self.esdl_asset.costInformation.marginalCosts is None:
            logger.warning(
                f"No marginal costs found for asset, Marginal costs set to 0 for: "
                f"{self.esdl_asset.name}"
            )
            return 0
        return float(self.esdl_asset.costInformation.marginalCosts.value)


def get_return_temperature(esdl_port: esdl.Port) -> float:
    """Get the temperature of the port."""
    return float(esdl_port.carrier.returnTemperature) + 273.15


def get_supply_temperature(esdl_port: esdl.Port) -> float:
    """Get the temperature of the port."""
    return float(esdl_port.carrier.supplyTemperature) + 273.15



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module containing Based boundary class."""
import numpy as np

from omotes_simulator_core.solver.matrix.equation_object import EquationObject
from omotes_simulator_core.solver.network.assets.base_asset import BaseAsset
from omotes_simulator_core.solver.matrix.index_core_quantity import index_core_quantity


class BaseBoundary(BaseAsset):
    """A base class for boundary conditions.

    This class represents a boundary condition and
    has a number of connection points, a name, and a number of unknowns.

    Attributes:
        connected_nodes (list): A list of nodes that are connected to this boundary.
        number_of_connection_point (int): The number of connection points for this boundary.
        equations_dict (dict): A dictionary that maps equation ids to equation indices in the
    """

    def __init__(
        self, name: str, _id: str, number_of_unknowns: int = 3, number_con_points: int = 1
    ):
        """Initialize the boundary condition.

        :param int number_con_points:The number of connection points for this boundary.
        :param str name: The name of this boundary.
        :param int number_of_unknowns (optional): The number of unknowns for this boundary.
                                                  Defaults to 3.
        """
        super().__init__(
            name=name,
            _id=_id,
            number_of_unknowns=number_of_unknowns,
            number_connection_points=number_con_points,
        )
        self.initial_pressure = 10000.0

    def get_pressure_equation(self) -> EquationObject:
        """Get a prescribed pressure equation for the boundary.

        This method gets an equation to the matrix that sets the pressure at the boundary
        to a fixed value.
        :return: EquationObject
            An EquationObject that contains the indices, coefficients, and right-hand side
            value of the equation.
        """
        # self.equations_dict["prescribe pressure"] = equation_id
        equation_object = EquationObject()
        equation_object.indices = np.array([self.matrix_index + index_core_quantity.pressure])
        equation_object.coefficients = np.array([1.0])
        equation_object.rhs = self.initial_pressure
        return equation_object

    def get_equations(self) -> list[EquationObject]:
        """Method to get all the equations of the boundary.

        These equations are:
        - Pressure balance at each connection point
        - Thermal balance at each connection point
        - Pre describe pressure equation

        :return:EquationObject
            An EquationObject that contains the indices, coefficients, and right-hand side
            value of the equation.
        """
        equations = [
            self.get_pressure_equation(),
            self.get_thermal_equations(0),
            self.get_press_to_node_equation(0),
        ]
        return equations



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""Module containing a matrix class to store the matrix and solve it using numpy."""
import numpy as np
import numpy.typing as npt
import scipy as sp
import csv
from omotes_simulator_core.solver.matrix.equation_object import EquationObject
from omotes_simulator_core.solver.matrix.index_core_quantity import index_core_quantity


class Matrix:
    """Class which stores the matrix and can be used to solve it."""

    num_unknowns: int = 0
    sol_new: npt.NDArray = np.array([], dtype=float)
    sol_old: npt.NDArray = np.array([], dtype=float)
    relative_convergence: float = 1e-6
    absolute_convergence: float = 1e-6

    def __init__(self) -> None:
        """Constructor of matrix class."""
        pass

    def add_unknowns(self, number_unknowns: int) -> int:
        """Method to add unknowns to the matrix.

        Adds unknowns to the matrix and returns the location of the unknowns in the solution.
        :param int number_unknowns: Number of unknowns to add
        :return: int Starting index of the added unknowns.
        """
        if number_unknowns < 1:
            raise ValueError("Number of unknowns should be at least 1.")
        self.num_unknowns += number_unknowns
        self.sol_new = np.concatenate([self.sol_new, np.ones(number_unknowns)])
        self.sol_old = np.concatenate([self.sol_old, np.zeros(number_unknowns)])

        return self.num_unknowns - number_unknowns

    def solve(self, equations: list[EquationObject], dump: bool = False) -> list[float]:
        """Method to solve the system of equation given in the matrix using sparse matrix solver.

        A sparse matrix solver is used, for this the coefficients, indices in the matrix are
        converted to numpy arrays. These arrays are then used to create a csc_matrix which can
        be solved by the sparse matrix solver of scipy.
        :param dump: if true it will dump the matrix to a csv file
        :param equations: list with the equations to solve.
        :return: list containing the solution of the system of equations.
        """
        self.verify_equations(equations)
        self.sol_old = self.sol_new
        coefficient_array = np.concatenate([equation.coefficients for equation in equations])
        column_index_array = np.concatenate([equation.indices for equation in equations])
        row_index_array = np.concatenate(
            [np.full((len(equations[i])), i) for i in range(len(equations))]
        )
        matrix = sp.sparse.csc_matrix(
            (coefficient_array, (row_index_array, column_index_array)),
            shape=(self.num_unknowns, self.num_unknowns),
        )
        rhs = sp.sparse.csc_matrix([[equation.rhs] for equation in equations])
        if dump:
            self.dump_matrix(matrix=matrix, rhs_array=rhs)
        self.sol_new = sp.sparse.linalg.spsolve(matrix, rhs)
        if np.isnan(self.sol_new).any():
            raise RuntimeError("Matrix is singular")
        result: list[float] = self.sol_new.tolist()
        return result

    def verify_equations(self, equations: list[EquationObject]) -> None:
        """Method to verify if the system of equations can be solved.

        This method checks if the number of equations supplied is equal to the number of unknowns.
        :param equations: list with the equations to verify.
        :return: None
        """
        if len(equations) > self.num_unknowns:
            raise ValueError(
                f"Too many equations supplied. Got {len(equations)} equations, "
                f"but number of unknowns is {self.num_unknowns}"
            )
        if len(equations) < self.num_unknowns:
            raise ValueError(
                f"Not enough equation supplied. Got {len(equations)} equations, "
                f"but number of unknowns is {self.num_unknowns}"
            )

    def is_converged(self) -> bool:
        """Returns true when the solution has converged and false when not.

        This method uses the np.allclose method to calculate if the solution is converged
        :return: Bool whether the solution has converged based on the given convergence criteria.
        """
        if self.num_unknowns == 0:
            raise ValueError("No unknowns have been added to the matrix.")
        return np.allclose(
            self.sol_new,
            self.sol_old,
            atol=self.absolute_convergence,
            rtol=self.relative_convergence,
        )

    def get_solution(self, index: int, number_of_unknowns: int) -> list[float]:
        """Method to get the solution of an asset.

        Returns the solution starting at the given index and for the number of
        unknowns supplied at the input.

        :param int index:Start position for which to supply the solution
        :param int number_of_unknowns: Number of elements for which the solution needs to
        be provided
        :return: list with teh solution
        """
        if index > self.num_unknowns:
            raise IndexError(
                f"Index ({index}) greater than number " f"of unknowns ({self.num_unknowns})"
            )
        if (index + number_of_unknowns) > self.num_unknowns:
            raise IndexError(
                f"Index ({index}) plus request number of elements "
                f"({number_of_unknowns}) greater than "
                f"number of unknowns {self.num_unknowns}"
            )
        result: list[float] = self.sol_new[index : index + number_of_unknowns].tolist()
        return result

    def dump_matrix(
        self,
        matrix: sp.sparse.csc_matrix,
        rhs_array: sp.sparse.csc_matrix,
        file_name: str = "dump.csv",
    ) -> None:
        """Method to dump the matrix to a csv file.

        :param matrix: Matrix to be printed to the file
        :param rhs_array: Right hand side to be printed to the file
        :param file_name: File name to dump the matrix in default=dump.csv
        :return:
        """
        with open(file_name, "w", newline="") as f:
            write = csv.writer(f)
            write.writerow(
                ["m", "P", "u"]
                * int(self.num_unknowns / index_core_quantity.number_core_quantities)
                + ["rhs"]
            )
            for row, rhs in zip(matrix.todense(), rhs_array.todense()):
                write.writerow(row.tolist()[0] + rhs.tolist()[0])

    def reset_solution(self) -> None:
        """Method to reset the solution to 1, so the new iteration can start."""
        self.sol_new = np.ones(self.num_unknowns)



#  Copyright (c) 2023. Deltares & TNO
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""Module containing BaseAsset class."""
import math
from typing import Dict

import numpy as np

from omotes_simulator_core.solver.matrix.equation_object import EquationObject
from omotes_simulator_core.solver.network.assets.base_item import BaseItem
from omotes_simulator_core.solver.network.assets.base_node_item import BaseNodeItem
from omotes_simulator_core.solver.utils.fluid_properties import fluid_props


class BaseAsset(BaseItem):
    """A base class for assets in a network.

    This class inherits from BaseItem and provides methods for connecting nodes, checking connection
    status, and adding thermal and pressure equations.
    """

    connected_nodes: Dict[int, BaseNodeItem]

    def __init__(
        self,
        name: str,
        _id: str,
        number_of_unknowns: int = 6,
        number_connection_points: int = 2,
        supply_temperature: float = 293.15,
    ):
        """Initializes the BaseAsset object with the given parameters.

        :param str name: The name of the asset.
        :param str _id: The unique identifier of the asset.
        :param int, optional number_of_unknowns: The number of unknown variables for the node.
        The default is 3.
        :param int, optional number_connection_points: The number of connection points for the
            asset. The default is 2, which corresponds to the inlet and outlet.
        """
        super().__init__(
            name=name,
            _id=_id,
            number_of_unknowns=number_of_unknowns,
            number_connection_points=number_connection_points,
        )
        self.supply_temperature = supply_temperature
        self.connected_nodes = {}

    def check_connection_point_valid(self, connection_point: int) -> bool:
        """Checks if the connection point is valid for the asset.

        :param connection_point: The index of the connection point to check.
        :type connection_point: int
        :return: True if the connection point is valid, otherwise raises an IndexError.
        :rtype: bool
        """
        if connection_point > self.number_of_connection_point - 1:
            raise IndexError(
                f"Asset {self.name} only has {self.number_of_connection_point - 1}. "
                f"{connection_point} is too high"
            )
        else:
            return True

    def connect_node(self, connection_point: int, node: BaseNodeItem) -> None:
        """Connects a node to a connection point of the asset.

        :param connection_point: The index of the connection point to connect.
        :type connection_point: int
        :param node: The node to connect.
        :type node: Node
        :raises ValueError: If the connection point is already connected to a node.
        """
        self.check_connection_point_valid(connection_point)
        if connection_point in self.connected_nodes:
            raise ValueError(
                f" connection point {connection_point}  of asset {self.name} "
                f" already connected to a node"
            )

        self.connected_nodes[connection_point] = node

    def disconnect_node(self, connection_point: int) -> None:
        """Disconnects a node from a connection point of the asset.

        :param connection_point: The index of the connection point to disconnect.
        :type connection_point: int
        :raises ValueError: If the connection point is not connected to a node.
        """
        self.check_connection_point_valid(connection_point)
        del self.connected_nodes[connection_point]

    def is_connected(self, connection_point: int) -> bool:
        """Checks if a connection point is connected to a node.

        :param connection_point: The index of the connection point to check.
        :type connection_point: int
        :return: True if the connection point is connected, False otherwise.
        :rtype: bool
        """
        if self.check_connection_point_valid(connection_point):
            return connection_point in self.connected_nodes
        else:
            return False

    def get_connected_node(self, connection_point: int) -> BaseNodeItem:
        """Checks if a connection point is connected to a node.

        :param connection_point: The index of the connection point to check.
        :type connection_point: int
        :return: True if the connection point is connected, False otherwise.
        :rtype: bool
        """
        self.check_connection_point_valid(connection_point)

        if self.is_connected(connection_point):
            return self.connected_nodes[connection_point]
        raise ValueError(str(connection_point) + " is not connected")

    def is_all_connected(self) -> bool:
        """Checks if all connection points are connected to nodes.

        :return: True if all connection points are connected, False otherwise.
        :rtype: bool
        """
        result = [self.is_connected(i) for i in range(self.number_of_connection_point)]
        return all(result)

    def get_thermal_equations(self, connection_point: int) -> EquationObject:
        """Gets a thermal equation for a connection point of the asset.

        :param connection_point: The index of the connection point to get the equation for.
        :type connection_point: int
        :return: An equation object representing the thermal equation.
        :rtype: EquationObject
        """
        if (
            self.prev_sol[
                self.get_index_matrix(
                    property_name="mass_flow_rate",
                    connection_point=connection_point,
                    use_relative_indexing=True,
                )
            ]
            > 0
        ):
            return self.get_prescribe_temp_equation(connection_point)
        else:
            return self.get_internal_energy_to_node_equation(connection_point)

    def get_prescribe_temp_equation(self, connection_point: int) -> EquationObject:
        """Gets a prescribed temperature equation for a connection point of the asset.

        :param connection_point: The index of the connection point to get the equation for.
        :type connection_point: int
        :return: An equation object representing the prescribed temperature equation.
        :rtype: EquationObject
        """
        if not self.is_connected(connection_point=connection_point):
            raise ValueError(
                f"Connection point {connection_point} of asset {self.name} is not connected to a"
                + " node."
            )
        equation_object = EquationObject()
        equation_object.indices = np.array(
            [
                self.get_index_matrix(
                    property_name="internal_energy",
                    connection_point=connection_point,
                    use_relative_indexing=False,
                )
            ]
        )
        equation_object.coefficients = np.array([1.0])
        equation_object.rhs = fluid_props.get_ie(self.supply_temperature)
        return equation_object

    def get_internal_energy_to_node_equation(self, connection_point: int) -> EquationObject:
        """Gets a temperature to node equation for a connection point of the asset.

        :param connection_point: The index of the connection point to get the equation for.
        :type connection_point: int
        :return: An equation object representing the temperature to node equation.
        :rtype: EquationObject
        """
        # Check if the connection point is connected to a node
        if not self.is_connected(connection_point=connection_point):
            raise ValueError(
                f"Connection point {connection_point} of asset {self.name} is not connected to a"
                + " node."
            )

        equation_object = EquationObject()
        equation_object.indices = np.array(
            [
                self.get_index_matrix(
                    property_name="internal_energy",
                    connection_point=connection_point,
                    use_relative_indexing=False,
                ),
                self.get_connected_node(connection_point).get_index_matrix(
                    property_name="internal_energy", use_relative_indexing=False
                ),
            ]
        )
        equation_object.coefficients = np.array([1.0, -1.0])
        equation_object.rhs = 0.0
        return equation_object

    def set_physical_properties(self, physical_properties: Dict[str, float]) -> None:
        """Method to set the physical properties of the asset.

        This method is implemented in the derived classes of this class.
        """

    def get_press_to_node_equation(self, connection_point: int) -> EquationObject:
        """Gets a pressure to node equation for a connection point of the asset.

        :param connection_point: The index of the connection point to get the equation for.
        :type connection_point: int
        :return: An equation object representing the pressure to node equation.
        :rtype: EquationObject
        """
        # Check if the connection point is connected to a node
        if not self.is_connected(connection_point=connection_point):
            raise ValueError(
                f"Connection point {connection_point} of asset {self.name} is not connected to a"
                + " node."
            )

        equation_object = EquationObject()
        equation_object.indices = np.array(
            [
                self.get_index_matrix(
                    property_name="pressure",
                    connection_point=connection_point,
                    use_relative_indexing=False,
                ),
                self.get_connected_node(connection_point).get_index_matrix(
                    property_name="pressure", use_relative_indexing=False
                ),
            ]
        )
        equation_object.coefficients = np.array([1.0, -1.0])
        equation_object.rhs = 0.0
        return equation_object

    def get_result(self) -> list[float]:
        """Method to get the result of the asset.

        :return: a list of the unknowns which are solved.
        """
        results = []
        for i in range(self.number_of_connection_point):
            for j in range(math.floor(self.number_of_unknowns / self.number_of_connection_point)):
                if j == 2:
                    results.append(
                        fluid_props.get_t(
                            self.prev_sol[
                                i
                                * math.floor(
                                    self.number_of_unknowns / self.number_of_connection_point
                                )
                                + j
                            ]
                        )
                    )
                else:
                    results.append(
                        self.prev_sol[
                            i
                            * math.floor(self.number_of_unknowns / self.number_of_connection_point)
                            + j
                        ]
                    )
        return results

    def get_equations(self) -> list[EquationObject]:
        """Method to get the equations of the asset.

        This method is implemented in the derived classes of this class.

        :return: A list of equation objects representing the equations of the asset.
        """
        return []

    def get_mass_flow_rate(self, connection_point: int) -> float:
        """Method to get the mass flow rate of a connection point.

        :param int connection_point: The connection point for which to get the mass flow rate.
        :return: The mass flow rate of the connection point.
        """
        return float(
            self.prev_sol[
                self.get_index_matrix(
                    property_name="mass_flow_rate",
                    connection_point=connection_point,
                    use_relative_indexing=True,
                )
            ]
        )

    def get_pressure(self, connection_point: int) -> float:
        """Method to get the pressure of a connection point.

        :param int connection_point: The connection point for which to get the pressure.
        :return: The pressure of the connection point.
        """
        return float(
            self.prev_sol[
                self.get_index_matrix(
                    property_name="pressure",
                    connection_point=connection_point,
                    use_relative_indexing=True,
                )
            ]
        )

    def get_temperature(self, connection_point: int) -> float:
        """Method to get the temperature of a connection point.

        :param int connection_point: The connection point for which to get the temperature.
        :return: The temperature of the connection point.
        """
        return fluid_props.get_t(
            self.prev_sol[
                self.get_index_matrix(
                    property_name="internal_energy",
                    connection_point=connection_point,
                    use_relative_indexing=True,
                )
            ]
        )
